@@ -12,6 +12,7 @@
 #define OSMO_CTRL_PORT_TRX	4236
 /* 4237 used by VTY interface */
 #define OSMO_CTRL_PORT_BTS	4238
+#define OSMO_CTRL_PORT_BSC_NEIGH 4248 /* osmo-bsc Neighbor Resloution Service */
 #define OSMO_CTRL_PORT_NITB_BSC	4249
 #define OSMO_CTRL_PORT_BSC_NAT	4250
 #define OSMO_CTRL_PORT_SGSN	4251
@@ -26,4 +27,7 @@
 #define OSMO_CTRL_PORT_CBC	4265
 /* 4266 used by D-GSM mDNS */
 #define OSMO_CTRL_PORT_MGW	4267
+#define OSMO_CTRL_PORT_SMLC	4272
+/* 4273 used by VTY interface */
+#define OSMO_CTRL_PORT_HNODEB	4274
 /* When adding/changing port numbers, keep docs and wiki in sync. See above. */
