@@ -28,12 +28,26 @@ enum osmo_bts_features {
 	BTS_FEAT_IPV6_NSVC,
 	BTS_FEAT_ACCH_REP,
 	BTS_FEAT_CCN, /* Is CCN supported by the cell? TS 44.060 sec 8.8.2 */
+	BTS_FEAT_VAMOS, /* Is the BTS VAMOS capable? */
+	BTS_FEAT_ABIS_OSMO_PCU, /* BTS supports forwarding data to PCUIF over IPA OML multiplex */
+	BTS_FEAT_BCCH_POWER_RED,
+	BTS_FEAT_DYN_TS_SDCCH8, /* Osmo Dynamic TS supports configured as SDCCH8 */
+	BTS_FEAT_ACCH_TEMP_OVP, /* FACCH/SACCH Temporary overpower */
 	_NUM_BTS_FEAT
 };
 
 extern const struct value_string osmo_bts_features_descs[];
 
-const char *osmo_bts_feature_name(enum osmo_bts_features feature);
+static inline const char *osmo_bts_features_desc(enum osmo_bts_features val)
+{ return get_value_string(osmo_bts_features_descs, val); }
+
+const char *osmo_bts_feature_name(enum osmo_bts_features feature)
+	OSMO_DEPRECATED("Use osmo_bts_features_desc() instead");
+
+extern const struct value_string osmo_bts_features_names[];
+
+static inline const char *osmo_bts_features_name(enum osmo_bts_features val)
+{ return get_value_string(osmo_bts_features_names, val); }
 
 static inline int osmo_bts_set_feature(struct bitvec *features, enum osmo_bts_features feature)
 {
