@@ -67,7 +67,8 @@ struct ph_rach_ind_param {
 	/* elements added on 2018-02-26 */
 	int8_t rssi;		/*!< RSSI of RACH indication */
 	uint16_t ber10k;	/*!< BER in units of 0.01% */
-	int16_t acc_delay_256bits;/* !< Burst TA Offset in 1/256th bits */
+	int16_t acc_delay_256bits;/*!< Burst TA Offset in 1/256th bits */
+	int16_t lqual_cb;	/*!< Link quality in centiBel */
 };
 
 /*! for PH-[UNIT]DATA.{req,ind} | PH-RTS.ind */
@@ -78,11 +79,12 @@ struct ph_data_param {
 	int8_t rssi;		/*!< RSSI of receivedindication */
 	uint16_t ber10k;	/*!< BER in units of 0.01% */
 	union {
-		int16_t ta_offs_qbits;	/* !< Burst TA Offset in quarter bits */
+		int16_t ta_offs_qbits;	/*!< Burst TA Offset in quarter bits */
 		int16_t ta_offs_256bits;/*!< timing advance offset (in 1/256th bits) */
 	};
-	int16_t lqual_cb;	/* !< Link quality in centiBel */
+	int16_t lqual_cb;	/*!< Link quality in centiBel */
 	enum osmo_ph_pres_info_type pdch_presence_info; /*!< Info regarding presence/validity of header and data parts */
+	uint8_t is_sub:1;	/*!< flags */
 };
 
 /*! for TCH.{req,ind} | TCH-RTS.ind */
@@ -92,7 +94,9 @@ struct ph_tch_param {
 	int8_t rssi;		/*!< RSSI of received indication */
 	uint8_t marker;		/*!< RTP Marker bit (speech onset indicator) */
 	uint16_t ber10k;	/*!< BER in units of 0.01% */
-	int16_t lqual_cb;	/* !< Link quality in centiBel */
+	int16_t lqual_cb;	/*!< Link quality in centiBel */
+	int16_t ta_offs_256bits;/*!< timing advance offset (in 1/256th bits) */
+	uint8_t is_sub:1;	/*!< flags */
 };
 
 /*! for PH-CONN.ind */
@@ -122,7 +126,7 @@ struct info_meas_ind_param {
 /*! for {ACTIVATE,DEACTIVATE,MODIFY} MPH-INFO.req */
 struct info_act_req_param {
 	uint8_t chan_nr;	/*!< Channel Number (Like RSL) */
-	uint8_t sacch_only;	/*!< \breif Only deactivate SACCH */
+	uint8_t sacch_only;	/*!< \brief Only deactivate SACCH */
 };
 
 /*! for {ACTIVATE,DEACTIVATE} MPH-INFO.cnf */
