@@ -14,4 +14,8 @@ enum osmo_gsm_sap {
 	SAP_BSSGP_LL,
 	SAP_BSSGP_NM,
 	SAP_BSSGP_PFM,
+
+	SAP_NS,
+
+	SAP_BSSGP_RIM,
 };
