@@ -99,14 +99,21 @@ enum node_type {
 	L_CS7_SCCPADDR_NODE,	/*!< SS7 SCCP Address */
 	L_CS7_SCCPADDR_GT_NODE,	/*!< SS7 SCCP Global Title */
 
-	L_CPU_SCHED_NODE,		/*!< CPU Sched related options node */
+	L_CPU_SCHED_NODE,	/*!< CPU Sched related options node */
 	L_NS_BIND_NODE,		/*!< NS bind node */
 	L_NS_NSE_NODE,		/*!< NS NSE node */
 	/*
 	 * When adding new nodes to the libosmocore project, these nodes can be
 	 * used to avoid ABI changes for unrelated projects.
 	 */
+	RESERVED1_NODE,		/*!< Reserved for later extensions */
+	RESERVED2_NODE,		/*!< Reserved for later extensions */
 	RESERVED3_NODE,		/*!< Reserved for later extensions */
+	RESERVED4_NODE,		/*!< Reserved for later extensions */
+	RESERVED5_NODE,		/*!< Reserved for later extensions */
+	RESERVED6_NODE,		/*!< Reserved for later extensions */
+	RESERVED7_NODE,		/*!< Reserved for later extensions */
+	RESERVED8_NODE,		/*!< Reserved for later extensions */
 
 	_LAST_OSMOVTY_NODE
 };
@@ -450,7 +457,7 @@ extern struct cmd_element config_exit_cmd;
 extern struct cmd_element config_help_cmd;
 extern struct cmd_element config_list_cmd;
 extern struct cmd_element config_end_cmd;
-char *host_config_file();
+const char *host_config_file(void);
 void host_config_set(const char *);
 
 char *osmo_asciidoc_escape(const char *inp);
@@ -476,4 +483,6 @@ extern const struct value_string vty_ref_gen_mode_desc[];
 int vty_dump_xml_ref_mode(FILE *stream, enum vty_ref_gen_mode mode);
 int vty_dump_xml_ref(FILE *stream) OSMO_DEPRECATED("Use vty_dump_xml_ref_mode() instead");
 
+int vty_cmd_range_match(const char *range, const char *str);
+
 /*! @} */
