@@ -35,6 +35,7 @@
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
+#include <limits.h>
 
 #include <arpa/inet.h>
 
@@ -53,6 +54,7 @@
 
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/rate_ctr.h>
+#include <osmocom/core/stat_item.h>
 #include <osmocom/core/select.h>
 #include <osmocom/core/counter.h>
 #include <osmocom/core/talloc.h>
@@ -82,20 +84,16 @@ static LLIST_HEAD(ctrl_lookup_helpers);
  *  \returns 1 on success; 0 in case of error */
 int ctrl_parse_get_num(vector vline, int i, long *num)
 {
-	char *token, *tmp;
+	char *token;
+	int64_t val;
 
 	if (i >= vector_active(vline))
 		return 0;
 	token = vector_slot(vline, i);
 
-	errno = 0;
-	if (token[0] == '\0')
+	if (osmo_str_to_int64(&val, token, 10, LONG_MIN, LONG_MAX))
 		return 0;
-
-	*num = strtol(token, &tmp, 10);
-	if (tmp[0] != '\0' || errno != 0)
-		return 0;
-
+	*num = (long)val;
 	return 1;
 }
 
@@ -156,7 +154,7 @@ int ctrl_cmd_send_trap(struct ctrl_handle *ctrl, const char *name, char *value)
 	if (!cmd)
 		return -ENOMEM;
 
-	cmd->id = "0"; /* It's a TRAP! */
+	cmd->id = CTRL_CMD_TRAP_ID; /* It's a TRAP! */
 	cmd->variable = (char *) name;
 	cmd->reply = value;
 	r = ctrl_cmd_send_to_all(ctrl, cmd);
@@ -221,6 +219,10 @@ int ctrl_cmd_handle(struct ctrl_handle *ctrl, struct ctrl_cmd *cmd,
 
 	if (cmd->type == CTRL_TYPE_SET_REPLY ||
 	    cmd->type == CTRL_TYPE_GET_REPLY) {
+		if (ctrl->reply_cb) {
+			ctrl->reply_cb(ctrl, cmd, data);
+			return CTRL_CMD_HANDLED;
+		}
 		if (strncmp(cmd->reply, "OK", 2) == 0) {
 			LOGP(DLCTRL, LOGL_DEBUG, "%s <%s> for %s is OK\n",
 			     get_value_string(ctrl_type_vals, cmd->type),
@@ -360,7 +362,7 @@ static int handle_control_read(struct osmo_fd * bfd)
 		 * nothing left to do now. */
 		return 0;
 	} else if (ret < 0) {
-		LOGP(DLCTRL, LOGL_ERROR, "Failed to parse ip access message: %d\n", ret);
+		LOGP(DLCTRL, LOGL_ERROR, "Failed to parse ip access message: %d (%s)\n", ret, strerror(-ret));
 		return 0;
 	}
 
@@ -389,20 +391,44 @@ int ctrl_handle_msg(struct ctrl_handle *ctrl, struct ctrl_connection *ccon, stru
 	struct ipaccess_head_ext *iph_ext;
 	int result;
 
+	if (msg->len < sizeof(*iph)) {
+		LOGP(DLCTRL, LOGL_ERROR, "The message is too short.\n");
+		return -EINVAL;
+	}
+	iph = (struct ipaccess_head *) msg->data;
+	if (iph->proto == IPAC_PROTO_IPACCESS) {
+		uint8_t msg_type = *(msg->l2h);
+		switch (msg_type) {
+		case IPAC_MSGT_PING:
+			if (ipa_ccm_send_pong(ccon->write_queue.bfd.fd) < 0)
+				LOGP(DLINP, LOGL_ERROR, "Cannot send PONG message. Reason: %s\n", strerror(errno));
+			break;
+		case IPAC_MSGT_PONG:
+			break;
+		case IPAC_MSGT_ID_ACK:
+			if (ipa_ccm_send_id_ack(ccon->write_queue.bfd.fd) < 0)
+				LOGP(DLINP, LOGL_ERROR, "Cannot send ID_ACK message. Reason: %s\n", strerror(errno));
+			break;
+		default:
+			LOGP(DLCTRL, LOGL_DEBUG, "Received unhandled IPACCESS protocol message of type 0x%x: %s\n",
+			     msg_type, msgb_hexdump(msg));
+			break;
+		}
+		return 0;
+	}
+	if (iph->proto != IPAC_PROTO_OSMO) {
+		LOGP(DLCTRL, LOGL_ERROR, "Protocol mismatch. Received protocol 0x%x message: %s\n",
+		     iph->proto, msgb_hexdump(msg));
+		return -EINVAL;
+	}
 	if (msg->len < sizeof(*iph) + sizeof(*iph_ext)) {
 		LOGP(DLCTRL, LOGL_ERROR, "The message is too short.\n");
 		return -EINVAL;
 	}
-
-	iph = (struct ipaccess_head *) msg->data;
-	if (iph->proto != IPAC_PROTO_OSMO) {
-		LOGP(DLCTRL, LOGL_ERROR, "Protocol mismatch. We got 0x%x\n", iph->proto);
-		return -EINVAL;
-	}
-
 	iph_ext = (struct ipaccess_head_ext *) iph->data;
 	if (iph_ext->proto != IPAC_PROTO_EXT_CTRL) {
-		LOGP(DLCTRL, LOGL_ERROR, "Extended protocol mismatch. We got 0x%x\n", iph_ext->proto);
+		LOGP(DLCTRL, LOGL_ERROR, "Extended protocol mismatch. Received protocol 0x%x message: %s\n",
+		     iph_ext->proto, msgb_hexdump(msg));
 		return -EINVAL;
 	}
 
@@ -461,8 +487,14 @@ static int control_write_cb(struct osmo_fd *bfd, struct msgb *msg)
 		control_close_conn(ccon);
 		return -EBADF;
 	}
-	if (rc != msg->len)
+	if (rc < 0) {
 		LOGP(DLCTRL, LOGL_ERROR, "Failed to write message to the CTRL connection.\n");
+		return 0;
+	}
+	if (rc < msg->len) {
+		msgb_pull(msg, rc);
+		return -EAGAIN;
+	}
 
 	return 0;
 }
@@ -486,6 +518,7 @@ struct ctrl_connection *osmo_ctrl_conn_alloc(void *ctx, void *data)
 	INIT_LLIST_HEAD(&ccon->def_cmds);
 
 	ccon->write_queue.bfd.data = data;
+	ccon->write_queue.bfd.fd = -1;
 	ccon->write_queue.write_cb = control_write_cb;
 	ccon->write_queue.read_cb = handle_control_read;
 
@@ -500,7 +533,7 @@ static int listen_fd_cb(struct osmo_fd *listen_bfd, unsigned int what)
 	char *name;
 
 
-	if (!(what & BSC_FD_READ))
+	if (!(what & OSMO_FD_READ))
 		return 0;
 
 	fd = accept(listen_bfd->fd, NULL, NULL);
@@ -530,7 +563,7 @@ static int listen_fd_cb(struct osmo_fd *listen_bfd, unsigned int what)
 	LOGP(DLCTRL, LOGL_INFO, "accept()ed new CTRL connection from %s\n", name);
 
 	ccon->write_queue.bfd.fd = fd;
-	ccon->write_queue.bfd.when = BSC_FD_READ;
+	ccon->write_queue.bfd.when = OSMO_FD_READ;
 
 	ret = osmo_fd_register(&ccon->write_queue.bfd);
 	if (ret < 0) {
@@ -560,12 +593,12 @@ static uint64_t get_rate_ctr_value(const struct rate_ctr *ctr, int intv, const c
 	}
 }
 
-static int get_rate_ctr_group_idx(const struct rate_ctr_group *ctrg, int intv, struct ctrl_cmd *cmd)
+static int get_rate_ctr_group_idx(struct rate_ctr_group *ctrg, int intv, struct ctrl_cmd *cmd)
 {
 	unsigned int i;
 	for (i = 0; i < ctrg->desc->num_ctr; i++) {
 		ctrl_cmd_reply_printf(cmd, "%s %"PRIu64";", ctrg->desc->ctr_desc[i].name,
-				      get_rate_ctr_value(&ctrg->ctr[i], intv, ctrg->desc->group_name_prefix));
+				      get_rate_ctr_value(rate_ctr_group_get_ctr(ctrg, i), intv, ctrg->desc->group_name_prefix));
 		if (!cmd->reply) {
 			cmd->reply = "OOM";
 			return CTRL_CMD_ERROR;
@@ -694,6 +727,100 @@ static int verify_rate_ctr(struct ctrl_cmd *cmd, const char *value, void *data)
 	return 0;
 }
 
+/* Expose all stat_item groups on CTRL, as read-only variables of the form:
+ * stat_item.(last|...).group_name.N.item_name
+ * stat_item.(last|...).group_name.by_name.group_idx_name.item_name
+ */
+CTRL_CMD_DEFINE_RO(stat_item, "stat_item *");
+static int get_stat_item(struct ctrl_cmd *cmd, void *data)
+{
+	char *dup;
+	char *saveptr;
+	char *value_type;
+	char *group_name;
+	char *group_idx_str;
+	char *item_name;
+	char *tmp;
+	int32_t val;
+	struct osmo_stat_item_group *statg;
+	const struct osmo_stat_item *stat_item;
+
+	/* cmd will be freed in control_if.c after handling here, so no need to free the dup string. */
+	dup = talloc_strdup(cmd, cmd->variable);
+	if (!dup) {
+		cmd->reply = "OOM";
+		return CTRL_CMD_ERROR;
+	}
+
+	/* Split off the first "stat_item." part */
+	tmp = strtok_r(dup, ".", &saveptr);
+	if (!tmp || strcmp(tmp, "stat_item") != 0)
+		goto format_error;
+
+	/* Split off the "last." part (validated further below) */
+	value_type = strtok_r(NULL, ".", &saveptr);
+	if (!value_type)
+		goto format_error;
+
+	/* Split off the "group_name." part */
+	group_name = strtok_r(NULL, ".", &saveptr);
+	if (!group_name)
+		goto format_error;
+
+	/* Split off the "N." part */
+	group_idx_str = strtok_r(NULL, ".", &saveptr);
+	if (!group_idx_str)
+		goto format_error;
+	if (strcmp(group_idx_str, "by_name") == 0) {
+		/* The index is not given by "N" but by "by_name.foo". Get the "foo" idx-name */
+		group_idx_str = strtok_r(NULL, ".", &saveptr);
+		statg = osmo_stat_item_get_group_by_name_idxname(group_name, group_idx_str);
+	} else {
+		int idx;
+		if (osmo_str_to_int(&idx, group_idx_str, 10, 0, INT_MAX))
+			goto format_error;
+		statg = osmo_stat_item_get_group_by_name_idx(group_name, idx);
+	}
+	if (!statg) {
+		cmd->reply = "Stat group with given name and index not found";
+		return CTRL_CMD_ERROR;
+	}
+
+	/* Split off the "item_name" part */
+	item_name = strtok_r(NULL, ".", &saveptr);
+	if (!item_name)
+		goto format_error;
+	stat_item = osmo_stat_item_get_by_name(statg, item_name);
+	if (!stat_item) {
+		cmd->reply = "No such stat item found";
+		return CTRL_CMD_ERROR;
+	}
+
+	tmp = strtok_r(NULL, "", &saveptr);
+	if (tmp) {
+		cmd->reply = "Garbage after stat item name";
+		return CTRL_CMD_ERROR;
+	}
+
+	if (!strcmp(value_type, "last"))
+		val = osmo_stat_item_get_last(stat_item);
+	else
+		goto format_error;
+
+	cmd->reply = talloc_asprintf(cmd, "%"PRIu32, val);
+	if (!cmd->reply) {
+		cmd->reply = "OOM";
+		return CTRL_CMD_ERROR;
+	}
+
+	return CTRL_CMD_REPLY;
+
+format_error:
+	cmd->reply = "Stat item must be of form stat_item.type.group_name.N.item_name,"
+		" e.g. 'stat_item.last.bsc.0.msc_num:connected'";
+	return CTRL_CMD_ERROR;
+}
+
 /* counter */
 CTRL_CMD_DEFINE(counter, "counter *");
 static int get_counter(struct ctrl_cmd *cmd, void *data)
@@ -779,6 +906,9 @@ static int ctrl_init(unsigned int node_count)
 		goto err;
 
 	ret = ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_rate_ctr);
+	if (ret)
+		goto err_vec;
+	ret = ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_stat_item);
 	if (ret)
 		goto err_vec;
 	ret = ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_counter);
