@@ -83,11 +83,6 @@
 /* nanoseconds per bit (504) */
 #define BIT_DURATION_NS (1000000000 / SUPERCHANNEL_LINERATE)
 
-struct gre_hdr {
-	uint16_t flags;
-	uint16_t ptype;
-} __attribute__ ((packed));
-
 static void free_bind(struct gprs_ns2_vc_bind *bind);
 static int fr_dlci_rx_cb(void *cb_data, struct msgb *msg);
 
@@ -135,7 +130,7 @@ static void free_vc(struct gprs_ns2_vc *nsvc)
 	nsvc->priv = NULL;
 }
 
-static void dump_vty(const struct gprs_ns2_vc_bind *bind, struct vty *vty, bool _stats)
+static void dump_vty(const struct gprs_ns2_vc_bind *bind, struct vty *vty, bool stats)
 {
 	struct priv_bind *priv;
 	struct gprs_ns2_vc *nsvc;
@@ -151,7 +146,7 @@ static void dump_vty(const struct gprs_ns2_vc_bind *bind, struct vty *vty, bool
 		osmo_fr_role_str(fr_link->role), priv->if_running ? "UP" : "DOWN", VTY_NEWLINE);
 
 	llist_for_each_entry(nsvc, &bind->nsvc, blist) {
-		vty_out(vty, "    NSVCI %05u: %s%s", nsvc->nsvci, gprs_ns2_ll_str(nsvc), VTY_NEWLINE);
+		ns2_vty_dump_nsvc(vty, nsvc, stats);
 	}
 
 	priv = bind->priv;
@@ -256,13 +251,13 @@ static int fr_netif_ofd_cb(struct osmo_fd *bfd, uint32_t what)
 	if (!(what & OSMO_FD_READ))
 		return 0;
 
-	msg = msgb_alloc(NS_ALLOC_SIZE, "Gb/NS/FR/GRE Rx");
+	msg = msgb_alloc(NS_ALLOC_SIZE, "Gb/NS/FR Rx");
 	if (!msg)
 		return -ENOMEM;
 
 	rc = recvfrom(bfd->fd, msg->data, NS_ALLOC_SIZE, 0, (struct sockaddr *)&sll, &sll_len);
 	if (rc < 0) {
-		LOGBIND(bind, LOGL_ERROR, "recv error %s during NS-FR-GRE recv\n", strerror(errno));
+		LOGBIND(bind, LOGL_ERROR, "recv error %s during NS-FR recv\n", strerror(errno));
 		goto out_err;
 	} else if (rc == 0) {
 		goto out_err;
@@ -347,7 +342,7 @@ static void enqueue_at_head(struct gprs_ns2_vc_bind *bind, struct msgb *msg)
 {
 	struct priv_bind *priv = bind->priv;
 	llist_add(&msg->list, &priv->backlog.list);
-	osmo_stat_item_inc(bind->statg->items[NS2_BIND_STAT_BACKLOG_LEN], 1);
+	osmo_stat_item_inc(osmo_stat_item_group_get_item(bind->statg, NS2_BIND_STAT_BACKLOG_LEN), 1);
 	osmo_timer_schedule(&priv->backlog.timer, 0, priv->backlog.retry_us);
 }
 
@@ -355,7 +350,7 @@ static void enqueue_at_tail(struct gprs_ns2_vc_bind *bind, struct msgb *msg)
 {
 	struct priv_bind *priv = bind->priv;
 	llist_add_tail(&msg->list, &priv->backlog.list);
-	osmo_stat_item_inc(bind->statg->items[NS2_BIND_STAT_BACKLOG_LEN], 1);
+	osmo_stat_item_inc(osmo_stat_item_group_get_item(bind->statg, NS2_BIND_STAT_BACKLOG_LEN), 1);
 	osmo_timer_schedule(&priv->backlog.timer, 0, priv->backlog.retry_us);
 }
 
@@ -439,7 +434,7 @@ static void fr_backlog_timer_cb(void *data)
 			llist_add(&msg->list, &priv->backlog.list);
 			break;
 		}
-		osmo_stat_item_dec(bind->statg->items[NS2_BIND_STAT_BACKLOG_LEN], 1);
+		osmo_stat_item_dec(osmo_stat_item_group_get_item(bind->statg, NS2_BIND_STAT_BACKLOG_LEN), 1);
 	}
 
 restart_timer:
@@ -601,14 +596,18 @@ static void mtu_change(struct gprs_ns2_vc_bind *bind, uint32_t mtu)
 	struct priv_bind *bpriv = bind->priv;
 	struct gprs_ns2_nse *nse;
 
+	/* 2 byte DLCI header */
+	if (mtu <= 2)
+		return;
+	mtu -= 2;
+
 	if (mtu == bind->mtu)
 		return;
 
 	LOGBIND(bind, LOGL_INFO, "MTU changed from %d to %d.\n",
-		bind->mtu, mtu);
+		bind->mtu + 2, mtu + 2);
 
-	/* 2 byte DLCI header */
-	bind->mtu = mtu - 2;
+	bind->mtu = mtu;
 	if (!bpriv->if_running)
 		return;
 
@@ -976,8 +975,9 @@ struct gprs_ns2_vc *gprs_ns2_fr_connect(struct gprs_ns2_vc_bind *bind,
 		goto err;
 	}
 
-	snprintf(idbuf, sizeof(idbuf), "%s-%s-DLCI%u-NSE%05u-NSVC%05u", gprs_ns2_lltype_str(nse->ll),
-		 bpriv->netif, dlci, nse->nsei, nsvci);
+	snprintf(idbuf, sizeof(idbuf), "NSE%05u-NSVC%05u-%s-%s-DLCI%u", nse->nsei, nsvci,
+		 gprs_ns2_lltype_str(nse->ll), bpriv->netif, dlci);
+	osmo_identifier_sanitize_buf(idbuf, NULL, '_');
 	nsvc = ns2_vc_alloc(bind, nse, true, GPRS_NS2_VC_MODE_BLOCKRESET, idbuf);
 	if (!nsvc)
 		goto err;
