@@ -610,6 +610,10 @@ const struct value_string abis_nm_obj_class_names[] = {
 	{ NM_OC_RADIO_CARRIER,	"RADIO-CARRIER" },
 	{ NM_OC_BASEB_TRANSC,	"BASEBAND-TRANSCEIVER" },
 	{ NM_OC_CHANNEL,	"CHANNEL" },
+	{ NM_OC_IPAC_E1_TRUNK,	"IPAC-E1-TRUNK" },
+	{ NM_OC_IPAC_E1_PORT,	"IPAC-E1-PORT" },
+	{ NM_OC_IPAC_E1_CHAN,	"IPAC-E1-CHAN" },
+	{ NM_OC_IPAC_CLK_MODULE,"IPAC-CLK-MODULE" },
 	{ NM_OC_BS11_ADJC,	"ADJC" },
 	{ NM_OC_BS11_HANDOVER,	"HANDOVER" },
 	{ NM_OC_BS11_PWR_CTRL,	"POWER-CONTROL" },
@@ -699,7 +703,7 @@ static const enum abis_nm_chan_comb chcomb4pchan[] = {
 	[GSM_PCHAN_UNKNOWN]	= 0xff,
 	[GSM_PCHAN_CCCH_SDCCH4_CBCH]	= NM_CHANC_BCCH_CBCH,
 	[GSM_PCHAN_SDCCH8_SACCH8C_CBCH] = NM_CHANC_SDCCH_CBCH,
-	[GSM_PCHAN_TCH_F_TCH_H_PDCH]	= NM_CHANC_OSMO_TCHFull_TCHHalf_PDCH,
+	[GSM_PCHAN_OSMO_DYN]	= NM_CHANC_OSMO_DYN,
 	/* FIXME: bounds check */
 };
 
