@@ -1,3 +1,4 @@
+/* Cell Broadcast Service Protocol (CBSP, 3GPP TS 48.049): Message encoding, decoding and reception */
 /*
  * Copyright (C) 2019  Harald Welte <laforge@gnumonks.org>
  *
@@ -129,13 +130,13 @@ static int encode_wperiod(uint32_t secs)
 	if (secs <= 10)
 		return secs;
 	if (secs <= 30)
-		return (secs-10)/2;
+		return 10 + (secs-10)/2;
 	if (secs <= 120)
-		return (secs-30)/5;
+		return 30 + (secs-30)/5;
 	if (secs <= 600)
-		return (secs-120)/10;
+		return 120 + (secs-120)/10;
 	if (secs <= 60*60)
-		return (secs-600)/30;
+		return 600 + (secs-600)/30;
 	osmo_cbsp_errstr = "warning period out of range";
 	return -1;
 }
@@ -176,12 +177,15 @@ static int cbsp_enc_write_repl(struct msgb *msg, const struct osmo_cbsp_write_re
 		}
 	} else {
 		int wperiod = encode_wperiod(in->u.emergency.warning_period);
+		uint8_t *cur;
 		if (wperiod < 0)
 			return -EINVAL;
 		msgb_tv_put(msg, CBSP_IEI_EMERG_IND, in->u.emergency.indicator);
 		msgb_tv16_put(msg, CBSP_IEI_WARN_TYPE, in->u.emergency.warning_type);
-		msgb_tlv_put(msg, CBSP_IEI_WARN_SEC_INFO, sizeof(in->u.emergency.warning_sec_info),
-			     in->u.emergency.warning_sec_info);
+		/* Tag + fixed length value! */
+		msgb_put_u8(msg, CBSP_IEI_WARN_SEC_INFO);
+		cur = msgb_put(msg, sizeof(in->u.emergency.warning_sec_info));
+		memcpy(cur, in->u.emergency.warning_sec_info, sizeof(in->u.emergency.warning_sec_info));
 		msgb_tv_put(msg, CBSP_IEI_WARNING_PERIOD, wperiod);
 	}
 	return 0;
@@ -348,7 +352,12 @@ static int cbsp_enc_reset_fail(struct msgb *msg, const struct osmo_cbsp_reset_fa
 /* 8.1.3.18a KEEP ALIVE */
 static int cbsp_enc_keep_alive(struct msgb *msg, const struct osmo_cbsp_keep_alive *in)
 {
-	msgb_tv_put(msg, CBSP_IEI_KEEP_ALIVE_REP_PERIOD, in->repetition_period);
+	int rperiod = encode_wperiod(in->repetition_period);
+	if (in->repetition_period > 120)
+		return -EINVAL;
+	if (rperiod < 0)
+		return -EINVAL;
+	msgb_tv_put(msg, CBSP_IEI_KEEP_ALIVE_REP_PERIOD, rperiod);
 	return 0;
 }
 
@@ -674,6 +683,7 @@ static int cbsp_dec_write_repl(struct osmo_cbsp_write_replace *out, const struct
 		out->u.cbs.category = *TLVP_VAL(tp, CBSP_IEI_CATEGORY);
 		out->u.cbs.rep_period = tlvp_val16be(tp, CBSP_IEI_REP_PERIOD);
 		out->u.cbs.num_bcast_req = tlvp_val16be(tp, CBSP_IEI_NUM_BCAST_REQ);
+		out->u.cbs.dcs = *TLVP_VAL(tp, CBSP_IEI_DCS);
 		num_of_pages = *TLVP_VAL(tp, CBSP_IEI_NUM_OF_PAGES);
 		if (num_of_pages < 1)
 			return -EINVAL;
@@ -1078,12 +1088,14 @@ static int cbsp_dec_reset_fail(struct osmo_cbsp_reset_failure *out, const struct
 static int cbsp_dec_keep_alive(struct osmo_cbsp_keep_alive *out, const struct tlv_parsed *tp,
 				struct msgb *in, void *ctx)
 {
+	uint8_t rperiod;
 	if (!TLVP_PRES_LEN(tp, CBSP_IEI_KEEP_ALIVE_REP_PERIOD, 1)) {
 		osmo_cbsp_errstr = "missing/short mandatory IE";
 		return -EINVAL;
 	}
 
-	out->repetition_period = *TLVP_VAL(tp, CBSP_IEI_KEEP_ALIVE_REP_PERIOD);
+	rperiod = *TLVP_VAL(tp, CBSP_IEI_KEEP_ALIVE_REP_PERIOD);
+	out->repetition_period = decode_wperiod(rperiod);
 	return 0;
 }
 
@@ -1456,12 +1468,7 @@ int osmo_cbsp_recv_buffered(void *ctx, int fd, struct msgb **rmsg, struct msgb *
 		}
 	}
 	/* else: complete message received */
-	rc = msgb_l2len(msg);
-	if (rc == 0) {
-		/* drop empty message */
-		rc = -EAGAIN;
-		goto discard_msg;
-	}
+	rc = msgb_length(msg);
 	if (tmp_msg)
 		*tmp_msg = NULL;
 	*rmsg = msg;
