@@ -25,6 +25,7 @@
  *
  */
 
+#include <osmocom/gsm/gsm48.h>
 #include <osmocom/gsm/gsm0480.h>
 #include <osmocom/gsm/gsm_utils.h>
 
@@ -49,12 +50,12 @@ const struct value_string gsm0480_op_code_names[] = {
 	{ GSM0480_OP_CODE_ERASE_SS,			"EraseSS" },
 	{ GSM0480_OP_CODE_ACTIVATE_SS,			"ActivateSS" },
 	{ GSM0480_OP_CODE_DEACTIVATE_SS,		"DeactivateSS" },
-	{ GSM0480_OP_CODE_INTERROGATE_SS,		"IngerrogateSS" },
+	{ GSM0480_OP_CODE_INTERROGATE_SS,		"InterrogateSS" },
 	{ GSM0480_OP_CODE_NOTIFY_SS,			"NotifySS" },
 	{ GSM0480_OP_CODE_REGISTER_PASSWORD,		"RegisterPassword" },
 	{ GSM0480_OP_CODE_GET_PASSWORD,			"GetPassword" },
 	{ GSM0480_OP_CODE_PROCESS_USS_DATA,		"ProcessUSSD" },
-	{ GSM0480_OP_CODE_FORWARD_CHECK_SS_IND,		"ForwardChecckSSind" },
+	{ GSM0480_OP_CODE_FORWARD_CHECK_SS_IND,		"ForwardCheckSSind" },
 	{ GSM0480_OP_CODE_PROCESS_USS_REQ,		"ProcessUssReq" },
 	{ GSM0480_OP_CODE_USS_REQUEST,			"UssRequest" },
 	{ GSM0480_OP_CODE_USS_NOTIFY,			"UssNotify" },
@@ -67,26 +68,6 @@ const struct value_string gsm0480_op_code_names[] = {
 	{ 0, NULL }
 };
 
-static inline unsigned char *msgb_wrap_with_TL(struct msgb *msgb, uint8_t tag)
-{
-	uint8_t *data = msgb_push(msgb, 2);
-
-	data[0] = tag;
-	data[1] = msgb->len - 2;
-	return data;
-}
-
-static inline unsigned char *msgb_push_TLV1(struct msgb *msgb, uint8_t tag,
-					    uint8_t value)
-{
-	uint8_t *data = msgb_push(msgb, 3);
-
-	data[0] = tag;
-	data[1] = 1;
-	data[2] = value;
-	return data;
-}
-
 /* wrap an invoke around it... the other way around
  *
  * 1.) Invoke Component tag
@@ -97,13 +78,13 @@ static inline unsigned char *msgb_push_TLV1(struct msgb *msgb, uint8_t tag,
 int gsm0480_wrap_invoke(struct msgb *msg, int op, int link_id)
 {
 	/* 3. operation */
-	msgb_push_TLV1(msg, GSM0480_OPERATION_CODE, op);
+	msgb_tlv1_push(msg, GSM0480_OPERATION_CODE, op);
 
 	/* 2. invoke id tag */
-	msgb_push_TLV1(msg, GSM0480_COMPIDTAG_INVOKE_ID, link_id);
+	msgb_tlv1_push(msg, GSM0480_COMPIDTAG_INVOKE_ID, link_id);
 
 	/* 1. component tag */
-	msgb_wrap_with_TL(msg, GSM0480_CTYPE_INVOKE);
+	msgb_push_tl(msg, GSM0480_CTYPE_INVOKE);
 
 	return 0;
 }
@@ -111,7 +92,7 @@ int gsm0480_wrap_invoke(struct msgb *msg, int op, int link_id)
 /* wrap the GSM 04.08 Facility IE around it */
 int gsm0480_wrap_facility(struct msgb *msg)
 {
-	msgb_wrap_with_TL(msg, GSM0480_IE_FACILITY);
+	msgb_push_tl(msg, GSM0480_IE_FACILITY);
 
 	return 0;
 }
@@ -122,7 +103,7 @@ struct msgb *gsm0480_create_unstructuredSS_Notify(int alertPattern, const char *
 	uint8_t *seq_len_ptr, *ussd_len_ptr, *data;
 	int len;
 
-	msg = msgb_alloc_headroom(1024, 128, "GSM 04.80");
+	msg = gsm0480_msgb_alloc_name("TS 04.80 USSD Notify");
 	if (!msg)
 		return NULL;
 
@@ -168,7 +149,7 @@ struct msgb *gsm0480_create_notifySS(const char *text)
 	if (len < 1 || len > 160)
 		return NULL;
 
-	msg = msgb_alloc_headroom(1024, 128, "GSM 04.80");
+	msg = gsm0480_msgb_alloc_name("TS 04.80 NotifySS");
 	if (!msg)
 		return NULL;
 
@@ -342,7 +323,7 @@ int gsm0480_decode_ussd_request(const struct gsm48_hdr *hdr, uint16_t len,
 	memset(&ss, 0, sizeof(ss));
 
 	if (len < sizeof(*hdr) + 2) {
-		LOGP(0, LOGL_DEBUG, "USSD Request is too short.\n");
+		LOGP(DLGLOBAL, LOGL_ERROR, "USSD Request is too short.\n");
 		return 0;
 	}
 
@@ -364,7 +345,7 @@ int gsm0480_decode_ussd_request(const struct gsm48_hdr *hdr, uint16_t len,
 	}
 
 	if (!rc)
-		LOGP(0, LOGL_DEBUG, "Error occurred while parsing received USSD!\n");
+		LOGP(DLGLOBAL, LOGL_ERROR, "Error occurred while parsing received USSD!\n");
 
 	return rc;
 }
@@ -381,7 +362,7 @@ int gsm0480_decode_ss_request(const struct gsm48_hdr *hdr, uint16_t len,
 	 */
 	pdisc = gsm48_hdr_pdisc(hdr);
 	if (pdisc != GSM48_PDISC_NC_SS) {
-		LOGP(0, LOGL_ERROR, "Dropping message with "
+		LOGP(DLGLOBAL, LOGL_ERROR, "Dropping message with "
 			"unsupported pdisc=%02x\n", pdisc);
 		return 0;
 	}
@@ -404,7 +385,7 @@ static int parse_ss(const struct gsm48_hdr *hdr, uint16_t len, struct ss_request
 	 */
 	if (msg_type != GSM0480_MTYPE_RELEASE_COMPLETE) {
 		if (len < 2) {
-			LOGP(0, LOGL_DEBUG, "SS Request is too short.\n");
+			LOGP(DLGLOBAL, LOGL_ERROR, "SS Request is too short.\n");
 			return 0;
 		}
 	}
@@ -412,8 +393,6 @@ static int parse_ss(const struct gsm48_hdr *hdr, uint16_t len, struct ss_request
 	/* Table 2.1: Messages for call independent SS control */
 	switch (msg_type) {
 	case GSM0480_MTYPE_RELEASE_COMPLETE:
-		LOGP(0, LOGL_DEBUG, "SS Release Complete\n");
-
 		/**
 		 * Indicates that there is no decoded message.
 		 * To be overwriten by the message otherwise.
@@ -431,7 +410,7 @@ static int parse_ss(const struct gsm48_hdr *hdr, uint16_t len, struct ss_request
 		rc &= parse_ss_facility(&hdr->data[0], len, req);
 		break;
 	default:
-		LOGP(0, LOGL_DEBUG, "Unknown GSM 04.80 message-type field 0x%02x\n",
+		LOGP(DLGLOBAL, LOGL_ERROR, "Unknown GSM 04.80 message-type field 0x%02x\n",
 			hdr->msg_type);
 		rc = 0;
 		break;
@@ -480,7 +459,7 @@ static int parse_ss_info_elements(const uint8_t *ss_ie, uint16_t len,
 	case GSM0480_IE_SS_VERSION:
 		break;
 	default:
-		LOGP(0, LOGL_DEBUG, "Unhandled GSM 04.08 or 04.80 IEI 0x%02x\n",
+		LOGP(DLGLOBAL, LOGL_ERROR, "Unhandled GSM 04.08 or 04.80 IEI 0x%02x\n",
 			iei);
 		rc = 0;
 		break;
@@ -516,7 +495,7 @@ int gsm0480_parse_facility_ie(const uint8_t *facility_ie, uint16_t length,
 
 		/* Make sure that there is no overflow */
 		if (offset + 2 + component_length > length) {
-			LOGP(0, LOGL_ERROR, "Component does not fit.\n");
+			LOGP(DLGLOBAL, LOGL_ERROR, "Component does not fit.\n");
 			return -EINVAL;
 		}
 
@@ -536,7 +515,7 @@ int gsm0480_parse_facility_ie(const uint8_t *facility_ie, uint16_t length,
 		case GSM0480_CTYPE_REJECT:
 			break;
 		default:
-			LOGP(0, LOGL_DEBUG, "Unknown GSM 04.80 Facility "
+			LOGP(DLGLOBAL, LOGL_ERROR, "Unknown GSM 04.80 Facility "
 				"Component Type 0x%02x\n", component_type);
 			rc = 0;
 			break;
@@ -566,7 +545,7 @@ static int parse_ss_invoke(const uint8_t *invoke_data, uint16_t length,
 
 	/* mandatory part */
 	if (invoke_data[0] != GSM0480_COMPIDTAG_INVOKE_ID) {
-		LOGP(0, LOGL_DEBUG, "Unexpected GSM 04.80 Component-ID tag "
+		LOGP(DLGLOBAL, LOGL_ERROR, "Unexpected GSM 04.80 Component-ID tag "
 		     "0x%02x (expecting Invoke ID tag)\n", invoke_data[0]);
 	}
 
@@ -608,13 +587,13 @@ static int parse_ss_invoke(const uint8_t *invoke_data, uint16_t length,
 						 req);
 			break;
 		default:
-			LOGP(0, LOGL_DEBUG, "GSM 04.80 operation code 0x%02x "
+			LOGP(DLGLOBAL, LOGL_ERROR, "GSM 04.80 operation code 0x%02x "
 				"is not yet handled\n", operation_code);
 			rc = 0;
 			break;
 		}
 	} else {
-		LOGP(0, LOGL_DEBUG, "Unexpected GSM 04.80 Component-ID tag 0x%02x "
+		LOGP(DLGLOBAL, LOGL_ERROR, "Unexpected GSM 04.80 Component-ID tag 0x%02x "
 			"(expecting Operation Code tag)\n",
 			invoke_data[0]);
 		rc = 0;
@@ -635,7 +614,7 @@ static int parse_ss_return_result(const uint8_t *rr_data, uint16_t length,
 
 	/* Mandatory part */
 	if (rr_data[0] != GSM0480_COMPIDTAG_INVOKE_ID) {
-		LOGP(0, LOGL_DEBUG, "Unexpected GSM 04.80 Component-ID tag "
+		LOGP(DLGLOBAL, LOGL_ERROR, "Unexpected GSM 04.80 Component-ID tag "
 		     "0x%02x (expecting Invoke ID tag)\n", rr_data[0]);
 		return 0;
 	}
@@ -666,7 +645,7 @@ static int parse_ss_return_result(const uint8_t *rr_data, uint16_t length,
 		return parse_process_uss_data(rr_data + offset + 3,
 			length - offset - 3, req);
 	default:
-		LOGP(0, LOGL_DEBUG, "GSM 04.80 operation code 0x%02x "
+		LOGP(DLGLOBAL, LOGL_ERROR, "GSM 04.80 operation code 0x%02x "
 			"is not yet handled\n", operation_code);
 		return 0;
 	}
@@ -787,13 +766,22 @@ static int parse_ss_for_bs_req(const uint8_t *ss_req_data,
 	return rc;
 }
 
-struct msgb *gsm0480_create_ussd_resp(uint8_t invoke_id, uint8_t trans_id, const char *text)
+struct msgb *gsm0480_msgb_alloc_name(const char *name)
+{
+	return msgb_alloc_headroom(1024, 128, name);
+}
+
+/*! Generate a USSD ReturnResult component containing a string in default GSM alphabet.
+ * \param[in] invoke_id		InvokeID of the request to which we respond
+ * \param[in] text		USSD text in ASCII; to be encoded as GSM 7-but alphabet
+ */
+struct msgb *gsm0480_gen_ussd_resp_7bit(uint8_t invoke_id, const char *text)
 {
 	struct msgb *msg;
 	uint8_t *ptr8;
 	int response_len;
 
-	msg = msgb_alloc_headroom(1024, 128, "GSM 04.80");
+	msg = gsm0480_msgb_alloc_name("TS 04.80 USSD Resp");
 	if (!msg)
 		return NULL;
 
@@ -803,46 +791,114 @@ struct msgb *gsm0480_create_ussd_resp(uint8_t invoke_id, uint8_t trans_id, const
 	msgb_put(msg, response_len);
 
 	/* Then wrap it as an Octet String */
-	msgb_wrap_with_TL(msg, ASN1_OCTET_STRING_TAG);
+	msgb_push_tl(msg, ASN1_OCTET_STRING_TAG);
 
 	/* Pre-pend the DCS octet string */
-	msgb_push_TLV1(msg, ASN1_OCTET_STRING_TAG, 0x0F);
+	msgb_tlv1_push(msg, ASN1_OCTET_STRING_TAG, 0x0F);
 
 	/* Then wrap these as a Sequence */
-	msgb_wrap_with_TL(msg, GSM_0480_SEQUENCE_TAG);
+	msgb_push_tl(msg, GSM_0480_SEQUENCE_TAG);
 
 	/* Pre-pend the operation code */
-	msgb_push_TLV1(msg, GSM0480_OPERATION_CODE,
+	msgb_tlv1_push(msg, GSM0480_OPERATION_CODE,
 			GSM0480_OP_CODE_PROCESS_USS_REQ);
 
 	/* Wrap the operation code and IA5 string as a sequence */
-	msgb_wrap_with_TL(msg, GSM_0480_SEQUENCE_TAG);
+	msgb_push_tl(msg, GSM_0480_SEQUENCE_TAG);
 
 	/* Pre-pend the invoke ID */
-	msgb_push_TLV1(msg, GSM0480_COMPIDTAG_INVOKE_ID, invoke_id);
+	msgb_tlv1_push(msg, GSM0480_COMPIDTAG_INVOKE_ID, invoke_id);
 
 	/* Wrap this up as a Return Result component */
-	msgb_wrap_with_TL(msg, GSM0480_CTYPE_RETURN_RESULT);
+	msgb_push_tl(msg, GSM0480_CTYPE_RETURN_RESULT);
 
-	/* Wrap the component in a Facility message */
-	msgb_wrap_with_TL(msg, GSM0480_IE_FACILITY);
-
-	/* And finally pre-pend the L3 header */
-	gsm0480_l3hdr_push(msg,
-			   GSM48_PDISC_NC_SS | trans_id
-			   | (1<<7) /* TI direction = 1 */,
-			   GSM0480_MTYPE_RELEASE_COMPLETE);
 	return msg;
 }
 
-struct gsm48_hdr *gsm0480_l3hdr_push(struct msgb *msg, uint8_t proto_discr,
-				     uint8_t msg_type)
+/*! Legacy helper: Generate USSD response including FACILITY IE + L3 header.
+ *
+ * This function is just like \ref gsm0480_gen_ussd_resp_7bit, but it generates
+ * not only the FACILITY value, but the full L3 message including message header
+ * and FACILITY IE Tag+Length.
+ */
+OSMO_DEPRECATED("Use gsm0480_gen_ussd_resp_7bit() instead")
+struct msgb *gsm0480_create_ussd_resp(uint8_t invoke_id, uint8_t trans_id, const char *text)
 {
-	struct gsm48_hdr *gh;
-	gh = (struct gsm48_hdr *) msgb_push(msg, sizeof(*gh));
-	gh->proto_discr = proto_discr;
-	gh->msg_type = msg_type;
-	return gh;
+	struct msgb *msg;
+
+	msg = gsm0480_gen_ussd_resp_7bit(invoke_id, text);
+	if (!msg)
+		return NULL;
+
+	/* Wrap the component in a Facility message */
+	msgb_push_tl(msg, GSM0480_IE_FACILITY);
+
+	/* And finally pre-pend the L3 header */
+	gsm48_push_l3hdr_tid(msg, GSM48_PDISC_NC_SS,
+			     /* FIXME: TI direction is always 1 ?!? */
+			     trans_id | (1 << 7),
+			     GSM0480_MTYPE_RELEASE_COMPLETE);
+
+	return msg;
+}
+
+/*! Generate a ReturnError component (see section 3.6.1) and given error code (see section 3.6.6).
+ * \param[in] invoke_id		InvokeID of the request
+ * \param[in] error_code	Error code (section 4.5)
+ * \return			message buffer containing the Reject component
+ */
+struct msgb *gsm0480_gen_return_error(uint8_t invoke_id, uint8_t error_code)
+{
+	struct msgb *msg;
+
+	msg = gsm0480_msgb_alloc_name("TS 04.80 ReturnError");
+	if (!msg)
+		return NULL;
+
+	/* First insert the problem code */
+	msgb_tlv1_push(msg, GSM_0480_ERROR_CODE_TAG, error_code);
+
+	/* Before it, insert the invoke ID */
+	msgb_tlv1_push(msg, GSM0480_COMPIDTAG_INVOKE_ID, invoke_id);
+
+	/* Wrap this up as a Reject component */
+	msgb_push_tl(msg, GSM0480_CTYPE_RETURN_ERROR);
+
+	/* FIXME: Wrap in Facility + L3? */
+	return msg;
+}
+
+/*! Generate a Reject component (see section 3.6.1) and given error code (see section 3.6.7).
+ * \param[in] invoke_id		InvokeID of the request
+ * \param[in] problem_tag	Problem code tag (table 3.13)
+ * \param[in] problem_code	Problem code (table 3.14-3.17)
+ * \return			message buffer containing the Reject component
+ *
+ * Note: if InvokeID is not available, e.g. when message parsing failed, any incorrect vlue
+ * can be passed (0x00 > x > 0xff), so the universal NULL-tag (see table 3.6) will be used instead.
+ */
+struct msgb *gsm0480_gen_reject(int invoke_id, uint8_t problem_tag, uint8_t problem_code)
+{
+	struct msgb *msg;
+
+	msg = gsm0480_msgb_alloc_name("TS 04.80 Reject");
+	if (!msg)
+		return NULL;
+
+	/* First insert the problem code */
+	msgb_tlv1_push(msg, problem_tag, problem_code);
+
+	/* If the Invoke ID is not available, Universal NULL (table 3.9) with length=0 shall be used */
+	if (invoke_id < 0 || invoke_id > 255)
+		msgb_tv_push(msg, ASN1_NULL_TYPE_TAG, 0);
+	else
+		msgb_tlv1_push(msg, GSM0480_COMPIDTAG_INVOKE_ID, invoke_id);
+
+	/* Wrap this up as a Reject component */
+	msgb_push_tl(msg, GSM0480_CTYPE_REJECT);
+
+	/* FIXME: Wrap in Facility + L3? */
+	return msg;
 }
 
 struct msgb *gsm0480_create_ussd_notify(int level, const char *text)
@@ -856,20 +912,47 @@ struct msgb *gsm0480_create_ussd_notify(int level, const char *text)
 	gsm0480_wrap_invoke(msg, GSM0480_OP_CODE_USS_NOTIFY, 0);
 	gsm0480_wrap_facility(msg);
 
-	gsm0480_l3hdr_push(msg, GSM48_PDISC_NC_SS, GSM0480_MTYPE_REGISTER);
+	/* And finally pre-pend the L3 header */
+	gsm48_push_l3hdr(msg, GSM48_PDISC_NC_SS,
+			 /* FIXME: no transactionID?!? */
+			 GSM0480_MTYPE_REGISTER);
+
 	return msg;
 }
 
+/*! Deprecated, use gsm0480_create_release_complete() instead. */
 struct msgb *gsm0480_create_ussd_release_complete(void)
 {
 	struct msgb *msg;
 
-	msg = msgb_alloc_headroom(1024, 128, "GSM 04.80 USSD REL COMPL");
+	msg = gsm0480_msgb_alloc_name("TS 04.80 USSD REL COMPL");
 	if (!msg)
 		return NULL;
 
-	/* FIXME: should this set trans_id and TI direction flag? */
-	gsm0480_l3hdr_push(msg, GSM48_PDISC_NC_SS,
-			   GSM0480_MTYPE_RELEASE_COMPLETE);
+	/* And finally pre-pend the L3 header */
+	gsm48_push_l3hdr(msg, GSM48_PDISC_NC_SS,
+			 /* FIXME: no transactionID?!? */
+			 GSM0480_MTYPE_RELEASE_COMPLETE);
+
+	return msg;
+}
+
+/*! Create a GSM 04.80 Release complete (see 2.5) message, prefixed
+ *  by GSM 04.08 L3 header with a given transaction ID.
+ * \param[in] trans_id  GSM 04.07 transaction identifier (and TI flag)
+ * \return  message buffer containing the Release complete message
+ */
+struct msgb *gsm0480_create_release_complete(uint8_t trans_id)
+{
+	struct msgb *msg;
+
+	msg = gsm0480_msgb_alloc_name("TS 04.80 USSD REL COMPL");
+	if (!msg)
+		return NULL;
+
+	/* Push the L3 header */
+	gsm48_push_l3hdr_tid(msg, GSM48_PDISC_NC_SS,
+			     trans_id, GSM0480_MTYPE_RELEASE_COMPLETE);
+
 	return msg;
 }
