@@ -721,6 +721,8 @@ uint8_t gsm0808_enc_encrypt_info(struct msgb *msg,
 	}
 
 	msgb_put_u8(msg, perm_algo);
+	/* FIXME: 48.008 3.2.2.10 Encryption Information says:
+	 * "When present, the key shall be 8 octets long." */
 	ptr = msgb_put(msg, ei->key_len);
 	memcpy(ptr, ei->key, ei->key_len);
 
@@ -760,6 +762,8 @@ int gsm0808_dec_encrypt_info(struct gsm0808_encrypt_info *ei,
 	}
 	ei->perm_algo_len = perm_algo_len;
 
+	/* FIXME: 48.008 3.2.2.10 Encryption Information says:
+	 * "When present, the key shall be 8 octets long." */
 	ei->key_len = len - 1;
 	memcpy(ei->key, elem, ei->key_len);
 	elem+=ei->key_len;
@@ -767,6 +771,30 @@ int gsm0808_dec_encrypt_info(struct gsm0808_encrypt_info *ei,
 	return (int)(elem - old_elem);
 }
 
+/*! Encode TS 48.008 Kc128 IE.
+ *  \param[out] msg  Message Buffer to which IE is to be appended.
+ *  \param[in] kc128  Pointer to 16 bytes of Kc128 key data.
+ *  \returns number of bytes appended to msg */
+int gsm0808_enc_kc128(struct msgb *msg, const uint8_t *kc128)
+{
+	uint8_t *start = msg->tail;
+	msgb_tv_fixed_put(msg, GSM0808_IE_KC_128, 16, kc128);
+	return msg->tail - start;
+}
+
+/*! Decode TS 48.008 Kc128 IE.
+ *  \param[out] kc128  Target buffer for received Kc128 key, 16 bytes long.
+ *  \param[in] elem  IE value to be decoded (without IE discriminator).
+ *  \param[in] len  Length of elem in bytes.
+ *  \returns number of bytes parsed; negative on error */
+int gsm0808_dec_kc128(uint8_t *kc128, const uint8_t *elem, uint8_t len)
+{
+	if (len != 16)
+		return -EINVAL;
+	memcpy(kc128, elem, 16);
+	return len;
+}
+
 /* Store individual Cell Identifier information in a CGI, without clearing the remaining ones.
  * This is useful to supplement one CGI with information from more than one Cell Identifier,
  * which in turn is useful to match Cell Identifiers of differing kinds to each other.
@@ -1579,7 +1607,7 @@ int gsm48_mr_cfg_from_gsm0808_sc_cfg(struct gsm48_multi_rate_conf *cfg,
 	/* Rate 5,15k can never be selected (see table) */
 	cfg->m5_15 = 0;
 
-	if (s15_s0 & GSM0808_SC_CFG_AMR_4_75_5_90_7_40_12_20 & 0xff) {
+	if (s15_s0 & GSM0808_SC_CFG_AMR_4_75_5_90_7_40_12_20) {
 		/* Table Table 7.11.3.1.3-2 lists one mode that selects 4
 		 * rates at once (Config-NB-Code = 1). The rates selected
 		 * are known to be compatible between GERAN and UTRAN, since
