@@ -98,15 +98,24 @@
 
 #include "../../config.h"
 
+#if (!EMBEDDED)
 /* FIXME: this can be removed once we bump glibc requirements to 2.25: */
-#if defined(__GLIBC__) && (__GLIBC__ >= 2) && (__GLIBC_MINOR__ >= 25)
+#ifdef __GLIBC_PREREQ
+#if __GLIBC_PREREQ(2,25)
+#define HAVE_GLIBC_GETRANDOM
+#endif /* if __GLIBC_PREREQ(2,25) */
+#endif /* ifdef __GLIBC_PREREQ */
+#ifdef HAVE_GLIBC_GETRANDOM
+#pragma message ("glibc " OSMO_STRINGIFY_VAL(__GLIBC__) "." OSMO_STRINGIFY_VAL(__GLIBC_MINOR__) " random detected")
 #include <sys/random.h>
+#undef USE_GNUTLS
 #elif HAVE_DECL_SYS_GETRANDOM
 #include <sys/syscall.h>
 #ifndef GRND_NONBLOCK
 #define GRND_NONBLOCK 0x0001
-#endif
-#endif
+#endif /* ifndef GRND_NONBLOCK */
+#endif /* ifdef HAVE_GLIBC_GETRANDOM */
+#endif /* !EMBEDDED */
 
 #if (USE_GNUTLS)
 #pragma message ("including GnuTLS for getrandom fallback.")
@@ -131,7 +140,8 @@ static void on_dso_unload_gnutls(void)
 	if (!gnutls_check_version("3.3.0"))
 		gnutls_global_deinit();
 }
-#endif
+
+#endif /* if (USE_GNUTLS) */
 
 /* ETSI GSM 03.38 6.2.1 and 6.2.1.1 default alphabet
  * Greek symbols at hex positions 0x10 and 0x12-0x1a
@@ -175,10 +185,10 @@ static int gsm_septet_lookup(uint8_t ch)
 	return -1;
 }
 
-/*! \brife Compute number of octets from number of septets,
- *  for instance: 47 septets needs 41,125 = 42 octets
- *  \param[in sept_len Number of Septets
- *  \returns Number of octets required */
+/*! Compute number of octets from number of septets.
+ * For instance: 47 septets need 41,125 = 42 octets.
+ * \param[in] sept_len Number of septets
+ * \returns Number of octets required */
 uint8_t gsm_get_octet_len(const uint8_t sept_len){
 	int octet_len = (sept_len * 7) / 8;
 	if ((sept_len * 7) % 8 != 0)
@@ -297,6 +307,7 @@ int gsm_septet_encode(uint8_t *result, const char *data)
 		case 0x5d:
 		case 0x7c:
 			result[y++] = 0x1b;
+		/* fall-through */
 		default:
 			result[y] = gsm_7bit_alphabet[ch];
 			break;
@@ -313,7 +324,7 @@ int gsm_septet_encode(uint8_t *result, const char *data)
  *  \param[in] septet_len Length of \a rdata
  *  \param[in] padding padding bits at start
  *  \returns number of bytes used in \a result */
-int gsm_septets2octets(uint8_t *result, const uint8_t *rdata, uint8_t septet_len, uint8_t padding)
+int gsm_septet_pack(uint8_t *result, const uint8_t *rdata, size_t septet_len, uint8_t padding)
 {
 	int i = 0, z = 0;
 	uint8_t cb, nb;
@@ -358,6 +369,12 @@ int gsm_septets2octets(uint8_t *result, const uint8_t *rdata, uint8_t septet_len
 	return z;
 }
 
+/*! Backwards compatibility wrapper for gsm_septets_pack(), deprecated. */
+int gsm_septets2octets(uint8_t *result, const uint8_t *rdata, uint8_t septet_len, uint8_t padding)
+{
+	return gsm_septet_pack(result, rdata, septet_len, padding);
+}
+
 /*! GSM 7-bit alphabet TS 03.38 6.2.1 Character packing
  *  \param[out] result Caller-provided output buffer
  *  \param[in] n Maximum length of \a result in bytes
@@ -382,7 +399,7 @@ int gsm_7bit_encode_n(uint8_t *result, size_t n, const char *data, int *octets)
 		y = max_septets;
 	}
 
-	o = gsm_septets2octets(result, rdata, y, 0);
+	o = gsm_septet_pack(result, rdata, y, 0);
 
 	if (octets)
 		*octets = o;
@@ -441,22 +458,23 @@ int osmo_get_rand_id(uint8_t *out, size_t len)
 	/* this function is intended for generating short identifiers only, not arbitrary-length random data */
 	if (len > OSMO_MAX_RAND_ID_LEN)
                return -E2BIG;
-
-#if defined(__GLIBC__) && (__GLIBC__ >= 2) && (__GLIBC_MINOR__ >= 25)
+#if (!EMBEDDED)
+#ifdef HAVE_GLIBC_GETRANDOM
 	rc = getrandom(out, len, GRND_NONBLOCK);
 #elif HAVE_DECL_SYS_GETRANDOM
 #pragma message ("Using direct syscall access for getrandom(): consider upgrading to glibc >= 2.25")
 	/* FIXME: this can be removed once we bump glibc requirements to 2.25: */
 	rc = syscall(SYS_getrandom, out, len, GRND_NONBLOCK);
 #endif
+#endif /* !EMBEDDED */
 
 	/* getrandom() failed entirely: */
 	if (rc < 0) {
 #if (USE_GNUTLS)
-#pragma message ("Secure random failed: using GnuTLS fallback.")
 		return gnutls_rnd(GNUTLS_RND_RANDOM, out, len);
-#endif
+#else
 		return -errno;
+#endif
 	}
 
 	/* getrandom() failed partially due to signal interruption:
@@ -475,7 +493,7 @@ int osmo_get_rand_id(uint8_t *out, size_t len)
  *  \param[out] buf Pre-allocated bufer for storing IE
  *  \returns Number of bytes filled in buf
  */
-size_t gsm0858_rsl_ul_meas_enc(struct gsm_meas_rep_unidir *mru, bool dtxd_used,
+size_t gsm0858_rsl_ul_meas_enc(const struct gsm_meas_rep_unidir *mru, bool dtxd_used,
 			uint8_t *buf)
 {
 	buf[0] = dtxd_used ? (1 << 6) : 0;
@@ -489,8 +507,8 @@ size_t gsm0858_rsl_ul_meas_enc(struct gsm_meas_rep_unidir *mru, bool dtxd_used,
 /*! Convert power class to dBm according to GSM TS 05.05
  *  \param[in] band GSM frequency band
  *  \param[in] class GSM power class
- *  \returns maximum transmit power of power class in dBm */
-unsigned int ms_class_gmsk_dbm(enum gsm_band band, int class)
+ *  \returns maximum transmit power of power class in dBm, negative on error */
+int ms_class_gmsk_dbm(enum gsm_band band, int class)
 {
 	switch (band) {
 	case GSM_BAND_450:
@@ -706,36 +724,61 @@ enum gsm_band gsm_band_parse(const char* mhz)
 	}
 }
 
-/*! Resolve GSM band from ARFCN
+/*! Resolve GSM band from ARFCN.
  *  In Osmocom, we use the highest bit of the \a arfcn to indicate PCS
  *  \param[in] arfcn Osmocom ARFCN, highest bit determines PCS mode
- *  \returns GSM Band */
-enum gsm_band gsm_arfcn2band(uint16_t arfcn)
+ *  \param[out] band GSM Band containing \arfcn if arfcn is valid, undetermined otherwise
+ *  \returns 0 if arfcn is valid and \a band was set, negative on error */
+int gsm_arfcn2band_rc(uint16_t arfcn, enum gsm_band *band)
 {
 	int is_pcs = arfcn & ARFCN_PCS;
 
 	arfcn &= ~ARFCN_FLAG_MASK;
 
-	if (is_pcs)
-		return GSM_BAND_1900;
-	else if (arfcn <= 124)
-		return GSM_BAND_900;
-	else if (arfcn >= 955 && arfcn <= 1023)
-		return GSM_BAND_900;
-	else if (arfcn >= 128 && arfcn <= 251)
-		return GSM_BAND_850;
-	else if (arfcn >= 512 && arfcn <= 885)
-		return GSM_BAND_1800;
-	else if (arfcn >= 259 && arfcn <= 293)
-		return GSM_BAND_450;
-	else if (arfcn >= 306 && arfcn <= 340)
-		return GSM_BAND_480;
-	else if (arfcn >= 350 && arfcn <= 425)
-		return GSM_BAND_810;
-	else if (arfcn >= 438 && arfcn <= 511)
-		return GSM_BAND_750;
-	else
-		return GSM_BAND_1800;
+	if (is_pcs) {
+		*band = GSM_BAND_1900;
+		return 0;
+	} else if (arfcn <= 124) {
+		*band = GSM_BAND_900;
+		return 0;
+	} else if (arfcn >= 955 && arfcn <= 1023) {
+		*band = GSM_BAND_900;
+		return 0;
+	} else if (arfcn >= 128 && arfcn <= 251) {
+		*band = GSM_BAND_850;
+		return 0;
+	} else if (arfcn >= 512 && arfcn <= 885) {
+		*band = GSM_BAND_1800;
+		return 0;
+	} else if (arfcn >= 259 && arfcn <= 293) {
+		*band = GSM_BAND_450;
+		return 0;
+	} else if (arfcn >= 306 && arfcn <= 340) {
+		*band = GSM_BAND_480;
+		return 0;
+	} else if (arfcn >= 350 && arfcn <= 425) {
+		*band = GSM_BAND_810;
+		return 0;
+	} else if (arfcn >= 438 && arfcn <= 511) {
+		*band = GSM_BAND_750;
+		return 0;
+	}
+	return -1;
+}
+
+/*! Resolve GSM band from ARFCN, aborts process on invalid ARFCN.
+ *  In Osmocom, we use the highest bit of the \a arfcn to indicate PCS.
+ *  DEPRECATED: Use gsm_arfcn2band_rc instead.
+ *  \param[in] arfcn Osmocom ARFCN, highest bit determines PCS mode
+ *  \returns GSM Band if ARFCN is valid (part of any valid band), aborts otherwise */
+enum gsm_band gsm_arfcn2band(uint16_t arfcn)
+{
+	enum gsm_band band;
+	if (gsm_arfcn2band_rc(arfcn, &band) == 0)
+		return band;
+
+	osmo_panic("%s:%d Invalid arfcn %" PRIu16 " passed to gsm_arfcn2band\n",
+		   __FILE__, __LINE__, arfcn);
 }
 
 struct gsm_freq_range {
@@ -851,14 +894,26 @@ uint32_t gsm_gsmtime2fn(struct gsm_time *time)
 	return (51 * ((time->t3 - time->t2 + 26) % 26) + time->t3 + (26 * 51 * time->t1));
 }
 
+char *osmo_dump_gsmtime_buf(char *buf, size_t buf_len, const struct gsm_time *tm)
+{
+	snprintf(buf, buf_len, "%06"PRIu32"/%02"PRIu16"/%02"PRIu8"/%02"PRIu8"/%02"PRIu8,
+		 tm->fn, tm->t1, tm->t2, tm->t3, (uint8_t)tm->fn%52);
+	buf[buf_len-1] = '\0';
+	return buf;
+}
+
 char *osmo_dump_gsmtime(const struct gsm_time *tm)
 {
-	static char buf[64];
+	static __thread char buf[64];
+	return osmo_dump_gsmtime_buf(buf, sizeof(buf), tm);
+}
 
-	snprintf(buf, sizeof(buf), "%06"PRIu32"/%02"PRIu16"/%02"PRIu8"/%02"PRIu8"/%02"PRIu8,
-		 tm->fn, tm->t1, tm->t2, tm->t3, (uint8_t)tm->fn%52);
-	buf[sizeof(buf)-1] = '\0';
-	return buf;
+char *osmo_dump_gsmtime_c(const void *ctx, const struct gsm_time *tm)
+{
+	char *buf = talloc_size(ctx, 64);
+	if (!buf)
+		return NULL;
+	return osmo_dump_gsmtime_buf(buf, 64, tm);
 }
 
 /*! append range1024 encoded data to bit vector
@@ -972,3 +1027,12 @@ int gsm_7bit_encode_oct(uint8_t *result, const char *data, int *octets)
 	return gsm_7bit_encode_n(result, GSM_7BIT_LEGACY_MAX_BUFFER_SIZE,
 				 data, octets);
 }
+
+/* This is also used by osmo-hlr's db schema */
+const struct value_string osmo_rat_type_names[] = {
+	{ OSMO_RAT_UNKNOWN, "unknown" },
+	{ OSMO_RAT_GERAN_A, "GERAN-A" },
+	{ OSMO_RAT_UTRAN_IU, "UTRAN-Iu" },
+	{ OSMO_RAT_EUTRAN_SGS, "EUTRAN-SGs" },
+	{}
+};
