@@ -126,6 +126,9 @@ const struct tlv_definition rsl_att_tlvdef = {
 		[RSL_IE_TFO_STATUS]		= { TLV_TYPE_TV },
 		[RSL_IE_LLP_APDU]		= { TLV_TYPE_TLV },
 		[RSL_IE_SIEMENS_MRPCI]		= { TLV_TYPE_TV },
+		[RSL_IE_OSMO_REP_ACCH_CAP]	= { TLV_TYPE_TLV },
+		[RSL_IE_OSMO_TRAINING_SEQUENCE]	= { TLV_TYPE_TLV },
+		[RSL_IE_OSMO_TEMP_OVP_ACCH_CAP]	= { TLV_TYPE_TLV },
 		[RSL_IE_IPAC_PROXY_UDP]		= { TLV_TYPE_FIXED, 2 },
 		[RSL_IE_IPAC_BSCMPL_TOUT]	= { TLV_TYPE_TV },
 		[RSL_IE_IPAC_REMOTE_IP]		= { TLV_TYPE_FIXED, 4 },
@@ -157,6 +160,7 @@ uint8_t rsl_enc_chan_nr(uint8_t type, uint8_t subch, uint8_t timeslot)
 
 	switch (type) {
 	case RSL_CHAN_Lm_ACCHs:
+	case RSL_CHAN_OSMO_VAMOS_Lm_ACCHs:
 		subch &= 0x01;
 		break;
 	case RSL_CHAN_SDCCH4_ACCH:
@@ -185,38 +189,34 @@ int rsl_dec_chan_nr(uint8_t chan_nr, uint8_t *type, uint8_t *subch, uint8_t *tim
 {
 	*timeslot = chan_nr & 0x7;
 
-	if ((chan_nr & 0xf8) == RSL_CHAN_Bm_ACCHs) {
-		*type = RSL_CHAN_Bm_ACCHs;
+	switch (chan_nr & RSL_CHAN_NR_MASK) {
+	case RSL_CHAN_Bm_ACCHs:
+	case RSL_CHAN_BCCH:
+	case RSL_CHAN_RACH:
+	case RSL_CHAN_PCH_AGCH:
+	case RSL_CHAN_OSMO_PDCH:
+	case RSL_CHAN_OSMO_CBCH4:
+	case RSL_CHAN_OSMO_CBCH8:
+	case RSL_CHAN_OSMO_VAMOS_Bm_ACCHs:
+		*type = chan_nr & RSL_CHAN_NR_MASK;
 		*subch = 0;
-	} else if ((chan_nr & 0xf0) == RSL_CHAN_Lm_ACCHs) {
-		*type = RSL_CHAN_Lm_ACCHs;
-		*subch = (chan_nr >> 3) & 0x1;
-	} else if ((chan_nr & 0xe0) == RSL_CHAN_SDCCH4_ACCH) {
-		*type = RSL_CHAN_SDCCH4_ACCH;
-		*subch = (chan_nr >> 3) & 0x3;
-	} else if ((chan_nr & 0xc0) == RSL_CHAN_SDCCH8_ACCH) {
-		*type = RSL_CHAN_SDCCH8_ACCH;
-		*subch = (chan_nr >> 3) & 0x7;
-	} else if ((chan_nr & 0xf8) == RSL_CHAN_BCCH) {
-		*type = RSL_CHAN_BCCH;
-		*subch = 0;
-	} else if ((chan_nr & 0xf8) == RSL_CHAN_RACH) {
-		*type = RSL_CHAN_RACH;
-		*subch = 0;
-	} else if ((chan_nr & 0xf8) == RSL_CHAN_PCH_AGCH) {
-		*type = RSL_CHAN_PCH_AGCH;
-		*subch = 0;
-	} else if ((chan_nr & 0xf8) == RSL_CHAN_OSMO_PDCH) {
-		*type = RSL_CHAN_OSMO_PDCH;
-		*subch = 0;
-	} else if ((chan_nr & 0xf8) == RSL_CHAN_OSMO_CBCH4) {
-		*type = RSL_CHAN_OSMO_CBCH4;
-		*subch = 0;
-	} else if ((chan_nr & 0xf8) == RSL_CHAN_OSMO_CBCH8) {
-		*type = RSL_CHAN_OSMO_CBCH8;
-		*subch = 0;
-	} else
-		return -EINVAL;
+		break;
+	default:
+		if ((chan_nr & 0xf0) == RSL_CHAN_Lm_ACCHs) {
+			*type = RSL_CHAN_Lm_ACCHs;
+			*subch = (chan_nr >> 3) & 0x1;
+		} else if ((chan_nr & 0xe0) == RSL_CHAN_SDCCH4_ACCH) {
+			*type = RSL_CHAN_SDCCH4_ACCH;
+			*subch = (chan_nr >> 3) & 0x3;
+		} else if ((chan_nr & 0xc0) == RSL_CHAN_SDCCH8_ACCH) {
+			*type = RSL_CHAN_SDCCH8_ACCH;
+			*subch = (chan_nr >> 3) & 0x7;
+		} else if ((chan_nr & 0xf0) == RSL_CHAN_OSMO_VAMOS_Lm_ACCHs) {
+			*type = RSL_CHAN_OSMO_VAMOS_Lm_ACCHs;
+			*subch = (chan_nr >> 3) & 0x1;
+		} else
+			return -EINVAL;
+	}
 
 	return 0;
 }
@@ -232,26 +232,30 @@ char *rsl_chan_nr_str_buf(char *buf, size_t buf_len, uint8_t chan_nr)
 	int ts = chan_nr & 7;
 	uint8_t cbits = chan_nr >> 3;
 
-	if (cbits == 0x01)
+	if (cbits == ABIS_RSL_CHAN_NR_CBITS_Bm_ACCHs)
 		snprintf(buf, buf_len, "TCH/F on TS%d", ts);
-	else if ((cbits & 0x1e) == 0x02)
+	else if ((cbits & 0x1e) == ABIS_RSL_CHAN_NR_CBITS_Lm_ACCHs(0))
 		snprintf(buf, buf_len, "TCH/H(%u) on TS%d", cbits & 0x01, ts);
-	else if ((cbits & 0x1c) == 0x04)
+	else if ((cbits & 0x1c) == ABIS_RSL_CHAN_NR_CBITS_SDCCH4_ACCH(0))
 		snprintf(buf, buf_len, "SDCCH/4(%u) on TS%d", cbits & 0x03, ts);
-	else if ((cbits & 0x18) == 0x08)
+	else if ((cbits & 0x18) == ABIS_RSL_CHAN_NR_CBITS_SDCCH8_ACCH(0))
 		snprintf(buf, buf_len, "SDCCH/8(%u) on TS%d", cbits & 0x07, ts);
-	else if (cbits == 0x10)
+	else if (cbits == ABIS_RSL_CHAN_NR_CBITS_BCCH)
 		snprintf(buf, buf_len, "BCCH on TS%d", ts);
-	else if (cbits == 0x11)
+	else if (cbits == ABIS_RSL_CHAN_NR_CBITS_RACH)
 		snprintf(buf, buf_len, "RACH on TS%d", ts);
-	else if (cbits == 0x12)
+	else if (cbits == ABIS_RSL_CHAN_NR_CBITS_PCH_AGCH)
 		snprintf(buf, buf_len, "PCH/AGCH on TS%d", ts);
-	else if (cbits == 0x18)
+	else if (cbits == ABIS_RSL_CHAN_NR_CBITS_OSMO_PDCH)
 		snprintf(buf, buf_len, "PDCH on TS%d", ts);
-	else if (cbits == 0x19)
+	else if (cbits == ABIS_RSL_CHAN_NR_CBITS_OSMO_CBCH4)
 		snprintf(buf, buf_len, "CBCH(SDCCH/4) on TS%d", ts);
-	else if (cbits == 0x1a)
+	else if (cbits == ABIS_RSL_CHAN_NR_CBITS_OSMO_CBCH8)
 		snprintf(buf, buf_len, "CBCH(SDCCH/8) on TS%d", ts);
+	else if (cbits == ABIS_RSL_CHAN_NR_CBITS_OSMO_VAMOS_Bm_ACCHs)
+		snprintf(buf, buf_len, "VAMOS TCH/F on TS%d", ts);
+	else if ((cbits & 0x1e) == ABIS_RSL_CHAN_NR_CBITS_OSMO_VAMOS_Lm_ACCHs(0))
+		snprintf(buf, buf_len, "VAMOS TCH/H(%u) on TS%d", cbits & 0x01, ts);
 	else
 		snprintf(buf, buf_len, "UNKNOWN on TS%d", ts);
 
@@ -264,7 +268,7 @@ char *rsl_chan_nr_str_buf(char *buf, size_t buf_len, uint8_t chan_nr)
  */
 const char *rsl_chan_nr_str(uint8_t chan_nr)
 {
-	static __thread char str[20];
+	static __thread char str[32];
 	return rsl_chan_nr_str_buf(str, sizeof(str), chan_nr);
 }
 
@@ -275,10 +279,10 @@ const char *rsl_chan_nr_str(uint8_t chan_nr)
  */
 char *rsl_chan_nr_str_c(const void *ctx, uint8_t chan_nr)
 {
-	char *str = talloc_size(ctx, 20);
+	char *str = talloc_size(ctx, 32);
 	if (!str)
 		return NULL;
-	return rsl_chan_nr_str_buf(str, 20, chan_nr);
+	return rsl_chan_nr_str_buf(str, 32, chan_nr);
 }
 
 static const struct value_string rsl_err_vals[] = {
@@ -617,6 +621,10 @@ const struct tlv_definition rsl_ipac_eie_tlvdef = {
 		[RSL_IPAC_EIE_3G_NCELL_LIST]	= { TLV_TYPE_TLV },
 		[RSL_IPAC_EIE_SDCCH_CTL_PARAM]	= { TLV_TYPE_TV },
 		[RSL_IPAC_EIE_AMR_CONV_THRESH] 	= { TLV_TYPE_FIXED, 9 },
+		/* Osmocom extensions: */
+		[RSL_IPAC_EIE_OSMO_MEAS_AVG_CFG]= { TLV_TYPE_TLV },
+		[RSL_IPAC_EIE_OSMO_MS_PWR_CTL]	= { TLV_TYPE_TLV },
+		[RSL_IPAC_EIE_OSMO_PC_THRESH_COMP]= { TLV_TYPE_TLV },
 	},
 };
 
