@@ -1,4 +1,4 @@
-/* (C) 2008-2017 by Harald Welte <laforge@gnumonks.org>
+/* (C) 2008-2020 by Harald Welte <laforge@gnumonks.org>
  * (C) 2016-2017 by sysmocom - s.f.m.c. GmbH
  *
  * All Rights Reserved
@@ -24,6 +24,7 @@
 #include <stdint.h>
 #include <errno.h>
 #include <osmocom/core/utils.h>
+#include <osmocom/core/logging.h>
 #include <osmocom/gsm/tlv.h>
 
 /*! \addtogroup tlv
@@ -234,7 +235,7 @@ int tlv_parse_one(uint8_t *o_tag, uint16_t *o_len, const uint8_t **o_val,
 	int len; /* number of bytes consumed by TLV entry */
 
 	if (buf_len < 1)
-		return -1;
+		return OSMO_TLVP_ERR_OFS_BEYOND_BUFFER;
 
 	tag = *buf;
 	*o_tag = tag;
@@ -268,7 +269,7 @@ int tlv_parse_one(uint8_t *o_tag, uint16_t *o_len, const uint8_t **o_val,
 	case TLV_TYPE_TLV:
 tlv:		/* GSM TS 04.07 11.2.4: Type 4 TLV */
 		if (buf_len < 2)
-			return -1;
+			return OSMO_TLVP_ERR_OFS_BEYOND_BUFFER;
 		*o_val = buf+2;
 		*o_len = *(buf+1);
 		len = *o_len + 2;
@@ -276,10 +277,10 @@ tlv:		/* GSM TS 04.07 11.2.4: Type 4 TLV */
 	case TLV_TYPE_vTvLV_GAN:	/* 44.318 / 11.1.4 */
 		/* FIXME: variable-length TAG! */
 		if (buf_len < 2)
-			return -1;
+			return OSMO_TLVP_ERR_OFS_BEYOND_BUFFER;
 		if (*(buf+1) & 0x80) {
 			if (buf_len < 3)
-				return -1;
+				return OSMO_TLVP_ERR_OFS_BEYOND_BUFFER;
 			/* like TL16Vbut without highest bit of len */
 			*o_val = buf+3;
 			*o_len = (*(buf+1) & 0x7F) << 8 | *(buf+2);
@@ -291,7 +292,7 @@ tlv:		/* GSM TS 04.07 11.2.4: Type 4 TLV */
 		break;
 	case TLV_TYPE_TvLV:
 		if (buf_len < 2)
-			return -1;
+			return OSMO_TLVP_ERR_OFS_BEYOND_BUFFER;
 		if (*(buf+1) & 0x80) {
 			/* like TLV, but without highest bit of len */
 			*o_val = buf+2;
@@ -302,18 +303,18 @@ tlv:		/* GSM TS 04.07 11.2.4: Type 4 TLV */
 		/* like TL16V, fallthrough */
 	case TLV_TYPE_TL16V:
 		if (buf_len < 3)
-			return -1;
+			return OSMO_TLVP_ERR_OFS_BEYOND_BUFFER;
 		*o_val = buf+3;
 		*o_len = *(buf+1) << 8 | *(buf+2);
 		len = *o_len + 3;
 		break;
 	default:
-		return -3;
+		return OSMO_TLVP_ERR_UNKNOWN_TLV_TYPE;
 	}
 
 	if (buf_len < len) {
 		*o_val = NULL;
-		return -2;
+		return OSMO_TLVP_ERR_OFS_LEN_BEYOND_BUFFER;
 	}
 	return len;
 }
@@ -370,12 +371,12 @@ int tlv_parse2(struct tlv_parsed *dec, int dec_multiples,
 		const uint8_t *val;
 		uint16_t parsed_len;
 		if (ofs > buf_len)
-			return -1;
+			return OSMO_TLVP_ERR_OFS_BEYOND_BUFFER;
 		val = &buf[ofs+1];
 		len = buf[ofs];
 		parsed_len = len + 1;
 		if (ofs + parsed_len > buf_len)
-			return -2;
+			return OSMO_TLVP_ERR_OFS_LEN_BEYOND_BUFFER;
 		num_parsed++;
 		ofs += parsed_len;
 		/* store the resulting val and len */
@@ -391,12 +392,12 @@ int tlv_parse2(struct tlv_parsed *dec, int dec_multiples,
 		const uint8_t *val;
 		uint16_t parsed_len;
 		if (ofs > buf_len)
-			return -1;
+			return OSMO_TLVP_ERR_OFS_BEYOND_BUFFER;
 		val = &buf[ofs+1];
 		len = buf[ofs];
 		parsed_len = len + 1;
 		if (ofs + parsed_len > buf_len)
-			return -2;
+			return OSMO_TLVP_ERR_OFS_LEN_BEYOND_BUFFER;
 		num_parsed++;
 		ofs += parsed_len;
 		/* store the resulting val and len */
@@ -432,7 +433,7 @@ int tlv_parse2(struct tlv_parsed *dec, int dec_multiples,
 	return num_parsed;
 }
 
-/*! take a master (src) tlvdev and fill up all empty slots in 'dst'
+/*! take a master (src) tlv_definition and fill up all empty slots in 'dst'
  *  \param dst TLV parser definition that is to be patched
  *  \param[in] src TLV parser definition whose content is patched into \a dst */
 void tlv_def_patch(struct tlv_definition *dst, const struct tlv_definition *src)
@@ -628,4 +629,108 @@ fail:
 	return -1;
 }
 
+static __thread char ienamebuf[32];
+static __thread char msgnamebuf[32];
+
+/*! get the message name for given msg_type in protocol pdef */
+const char *osmo_tlv_prot_msg_name(const struct osmo_tlv_prot_def *pdef, uint8_t msg_type)
+{
+	if (pdef->msg_def[msg_type].name) {
+		return pdef->msg_def[msg_type].name;
+	} else if (pdef->msgt_names) {
+		return get_value_string(pdef->msgt_names, msg_type);
+	} else {
+		snprintf(msgnamebuf, sizeof(msgnamebuf), "Unknown msg_type 0x%02x", msg_type);
+		return msgnamebuf;
+	}
+}
+
+/*! get the IE name for given IEI in protocol pdef */
+const char *osmo_tlv_prot_ie_name(const struct osmo_tlv_prot_def *pdef, uint8_t iei)
+{
+	if (pdef->ie_def[iei].name) {
+		return pdef->ie_def[iei].name;
+	} else {
+		snprintf(ienamebuf, sizeof(ienamebuf), "Unknown IEI 0x%02x", iei);
+		return ienamebuf;
+	}
+}
+
+/*! Validate an already TLV-decoded message against the protocol definition.
+ *  \param[in] pdef protocol definition of given protocol
+ *  \param[in] msg_type message type of the parsed message
+ *  \param[in] tp TLV parser result
+ *  \param[in] log_subsys logging sub-system for log messages
+ *  \param[in] log_pfx prefix for log messages
+ *  \returns 0 in case of success; negative osmo_tlv_parser_error in case of error
+ */
+int osmo_tlv_prot_validate_tp(const struct osmo_tlv_prot_def *pdef, uint8_t msg_type,
+			      const struct tlv_parsed *tp, int log_subsys, const char *log_pfx)
+{
+	const struct osmo_tlv_prot_msg_def *msg_def= &pdef->msg_def[msg_type];
+	unsigned int err = 0;
+	unsigned int i;
+
+	if (msg_def->mand_ies) {
+		for (i = 0; i < msg_def->mand_count; i++) {
+			uint8_t iei = msg_def->mand_ies[i];
+			if (!TLVP_PRESENT(tp, iei)) {
+				LOGP(log_subsys, LOGL_ERROR, "%s %s %s: Missing Mandatory IE: %s\n",
+				     log_pfx, pdef->name, osmo_tlv_prot_msg_name(pdef, msg_type),
+				     osmo_tlv_prot_ie_name(pdef, iei));
+				if (!err)
+					err = OSMO_TLVP_ERR_MAND_IE_MISSING;
+			}
+		}
+	}
+
+	for (i = 0; i < ARRAY_SIZE(tp->lv); i++) {
+		uint16_t min_len;
+
+		if (!TLVP_PRESENT(tp, i))
+			continue;
+
+		min_len = pdef->ie_def[i].min_len;
+		if (TLVP_LEN(tp, i) < min_len) {
+			LOGP(log_subsys, LOGL_ERROR, "%s %s %s: Short IE %s: %u < %u\n", log_pfx,
+			     pdef->name, osmo_tlv_prot_msg_name(pdef, msg_type),
+			     osmo_tlv_prot_ie_name(pdef, i), TLVP_LEN(tp, i), min_len);
+			if (!err)
+				err = OSMO_TLVP_ERR_IE_TOO_SHORT;
+		}
+	}
+
+	return err;
+}
+
+/*! Parse + Validate a TLV-encoded message against the protocol definition.
+ *  \param[in] pdef protocol definition of given protocol
+ *  \param[out] dec caller-allocated pointer to \ref tlv_parsed
+ *  \param[in] dec_multiples length of the tlv_parsed[] in \a dec.
+ *  \param[in] msg_type message type of the parsed message
+ *  \param[in] buf the input data buffer to be parsed
+ *  \param[in] buf_len length of the input data buffer
+ *  \param[in] lv_tag an initial LV tag at the start of the buffer
+ *  \param[in] lv_tag2 a second initial LV tag following the \a lv_tag
+ *  \param[in] log_subsys logging sub-system for log messages
+ *  \param[in] log_pfx prefix for log messages
+ *  \returns 0 in case of success; negative osmo_tlv_parser_error in case of error
+ */
+int osmo_tlv_prot_parse(const struct osmo_tlv_prot_def *pdef,
+			struct tlv_parsed *dec, unsigned int dec_multiples, uint8_t msg_type,
+			const uint8_t *buf, unsigned int buf_len, uint8_t lv_tag, uint8_t lv_tag2,
+			int log_subsys, const char *log_pfx)
+{
+	int rc;
+
+	rc = tlv_parse2(dec, dec_multiples, pdef->tlv_def, buf, buf_len, lv_tag, lv_tag2);
+	if (rc < 0) {
+		LOGP(log_subsys, LOGL_ERROR, "%s %s %s: TLV parser error %d\n", log_pfx,
+		     pdef->name, osmo_tlv_prot_msg_name(pdef, msg_type), rc);
+		return rc;
+	}
+
+	return osmo_tlv_prot_validate_tp(pdef, msg_type, dec, log_subsys, log_pfx);
+}
+
 /*! @} */
