@@ -90,13 +90,25 @@
 #include <osmocom/core/counter.h>
 #include <osmocom/core/msgb.h>
 
+#ifdef HAVE_SYSTEMTAP
+/* include the generated probes header and put markers in code */
+#include "probes.h"
+#define TRACE(probe) probe
+#define TRACE_ENABLED(probe) probe ## _ENABLED()
+#else
+/* Wrap the probe to allow it to be removed when no systemtap available */
+#define TRACE(probe)
+#define TRACE_ENABLED(probe) (0)
+#endif /* HAVE_SYSTEMTAP */
+
+#include <stat_item_internal.h>
+
 #define STATS_DEFAULT_INTERVAL 5 /* secs */
 #define STATS_DEFAULT_BUFLEN 256
 
-static LLIST_HEAD(osmo_stats_reporter_list);
+LLIST_HEAD(osmo_stats_reporter_list);
 static void *osmo_stats_ctx = NULL;
 static int is_initialised = 0;
-static int32_t current_stat_item_index = 0;
 
 static struct osmo_stats_config s_stats_config = {
 	.interval = STATS_DEFAULT_INTERVAL,
@@ -211,7 +223,7 @@ struct osmo_stats_reporter *osmo_stats_reporter_alloc(enum osmo_stats_reporter_t
 		srep->name = talloc_strdup(srep, name);
 	srep->fd = -1;
 
-	llist_add(&srep->list, &osmo_stats_reporter_list);
+	llist_add_tail(&srep->list, &osmo_stats_reporter_list);
 
 	return srep;
 }
@@ -225,13 +237,11 @@ void osmo_stats_reporter_free(struct osmo_stats_reporter *srep)
 	talloc_free(srep);
 }
 
-/*! Initilize the stats reporting module; call this once in your program
+/*! Initialize the stats reporting module; call this once in your program.
  *  \param[in] ctx Talloc context from which stats related memory is allocated */
 void osmo_stats_init(void *ctx)
 {
 	osmo_stats_ctx = ctx;
-	osmo_stat_item_discard_all(&current_stat_item_index);
-
 	is_initialised = 1;
 	start_timer();
 }
@@ -683,36 +693,28 @@ static int osmo_stat_item_handler(
 	struct osmo_stat_item_group *statg, struct osmo_stat_item *item, void *sctx_)
 {
 	struct osmo_stats_reporter *srep;
-	int32_t idx = current_stat_item_index;
-	int32_t value;
-	int have_value;
+	int32_t prev_reported_value = item->reported.max;
+	int32_t new_value = item->value.max;
 
-	have_value = osmo_stat_item_get_next(item, &idx, &value) > 0;
-	if (!have_value)
-		/* Send the last value in case a flush is requested */
-		value = osmo_stat_item_get_last(item);
+	llist_for_each_entry(srep, &osmo_stats_reporter_list, list) {
+		if (!srep->running)
+			continue;
 
-	do {
-		llist_for_each_entry(srep, &osmo_stats_reporter_list, list) {
-			if (!srep->running)
-				continue;
+		/* If the previously reported value is the same as the current value, skip resending the value.
+		 * However, if the stats reporter is set to resend all values, do resend the current value regardless of
+		 * repetitions.
+		 */
+		if (new_value == prev_reported_value && !srep->force_single_flush)
+			continue;
 
-			if (!have_value && !srep->force_single_flush)
-				continue;
+		if (!osmo_stats_reporter_check_config(srep,
+				statg->idx, statg->desc->class_id))
+			continue;
 
-			if (!osmo_stats_reporter_check_config(srep,
-					statg->idx, statg->desc->class_id))
-				continue;
+		osmo_stats_reporter_send_item(srep, statg, item->desc, new_value);
+	}
 
-			osmo_stats_reporter_send_item(srep, statg,
-				item->desc, value);
-		}
-
-		if (!have_value)
-			break;
-
-		have_value = osmo_stat_item_get_next(item, &idx, &value) > 0;
-	} while (have_value);
+	osmo_stat_item_flush(item);
 
 	return 0;
 }
@@ -781,13 +783,14 @@ static void flush_all_reporters()
 int osmo_stats_report()
 {
 	/* per group actions */
+	TRACE(LIBOSMOCORE_STATS_START());
 	osmo_counters_for_each(handle_counter, NULL);
 	rate_ctr_for_each_group(rate_ctr_group_handler, NULL);
 	osmo_stat_item_for_each_group(osmo_stat_item_group_handler, NULL);
 
 	/* global actions */
-	osmo_stat_item_discard_all(&current_stat_item_index);
 	flush_all_reporters();
+	TRACE(LIBOSMOCORE_STATS_DONE());
 
 	return 0;
 }
