@@ -82,7 +82,7 @@ static struct sched_vty_opts *sched_vty_opts;
 
 static struct cmd_node sched_node = {
 	L_CPU_SCHED_NODE,
-	"%s(config-sched)# ",
+	"%s(config-cpu-sched)# ",
 	1,
 };
 
@@ -175,7 +175,7 @@ static int generate_cpu_hex_mask(char *str, size_t str_buf_size,
 		if (CPU_ISSET_S(cpu + 3, cpuset_size, cpuset))
 			val |= 0x08;
 
-		if (val >= 0 && val < 10)
+		if (val < 10)
 			*ptr = '0' + val;
 		else
 			*ptr = ('a' - 10) + val;
@@ -276,7 +276,6 @@ static bool proc_name_exists(const char *name, pid_t *res_pid)
 static enum sched_vty_thread_id procname2pid(pid_t *res_pid, const char *str, bool applynow)
 {
 	size_t i, len;
-	char *end;
 	bool is_pid = true;
 
 	if (strcmp(str, "all") == 0) {
@@ -297,12 +296,12 @@ static enum sched_vty_thread_id procname2pid(pid_t *res_pid, const char *str, bo
 		}
 	}
 	if (is_pid) {
-		errno = 0;
-		*res_pid = strtoul(str, &end, 0);
-		if ((errno == ERANGE && *res_pid == ULONG_MAX) || (errno && !*res_pid) ||
-		    str == end) {
+		int64_t val;
+		if (osmo_str_to_int64(&val, str, 0, 0, INT64_MAX))
+			return SCHED_VTY_THREAD_UNKNOWN;
+		*res_pid = (pid_t)val;
+		if (*res_pid != val)
 			return SCHED_VTY_THREAD_UNKNOWN;
-		}
 		if (!applynow || proc_tid_exists(*res_pid))
 			return SCHED_VTY_THREAD_ID;
 		else
@@ -371,7 +370,7 @@ static int my_sched_setaffinity(enum sched_vty_thread_id tid_type, pid_t pid,
 
 }
 
-DEFUN(cfg_sched_cpu_affinity, cfg_sched_cpu_affinity_cmd,
+DEFUN_ATTR(cfg_sched_cpu_affinity, cfg_sched_cpu_affinity_cmd,
 	"cpu-affinity (self|all|<0-4294967295>|THREADNAME) CPUHEXMASK [delay]",
 	"Set CPU affinity mask on a (group of) thread(s)\n"
 	"Set CPU affinity mask on thread running the VTY\n"
@@ -379,7 +378,8 @@ DEFUN(cfg_sched_cpu_affinity, cfg_sched_cpu_affinity_cmd,
 	"Set CPU affinity mask on a thread with specified PID\n"
 	"Set CPU affinity mask on a thread with specified thread name\n"
 	"CPU affinity mask\n"
-	"If set, delay applying the affinity mask now and let the app handle it at a later point\n")
+	"If set, delay applying the affinity mask now and let the app handle it at a later point\n",
+	CMD_ATTR_IMMEDIATE)
 {
 	const char* str_who = argv[0];
 	const char *str_mask = argv[1];
@@ -469,18 +469,19 @@ static int set_sched_rr(unsigned int prio)
 	LOGP(DLGLOBAL, LOGL_NOTICE, "Setting SCHED_RR priority %d\n", param.sched_priority);
 	rc = sched_setscheduler(getpid(), SCHED_RR, &param);
 	if (rc == -1) {
-		LOGP(DLGLOBAL, LOGL_FATAL, "Setting SCHED_RR priority %d failed: %s\n",
+		LOGP(DLGLOBAL, LOGL_ERROR, "Setting SCHED_RR priority %d failed: %s\n",
 		     param.sched_priority, strerror(errno));
 		return -1;
 	}
 	return 0;
 }
 
-DEFUN(cfg_sched_policy, cfg_sched_policy_cmd,
+DEFUN_ATTR(cfg_sched_policy, cfg_sched_policy_cmd,
 	"policy rr <1-32>",
 	"Set the scheduling policy to use for the process\n"
 	"Use the SCHED_RR real-time scheduling algorithm\n"
-	"Set the SCHED_RR real-time priority\n")
+	"Set the SCHED_RR real-time priority\n",
+	CMD_ATTR_IMMEDIATE)
 {
 	sched_vty_opts->sched_rr_prio = atoi(argv[0]);
 
@@ -604,13 +605,13 @@ int osmo_cpu_sched_vty_init(void *tall_ctx)
 	INIT_LLIST_HEAD(&sched_vty_opts->cpu_affinity_li);
 	pthread_mutex_init(&sched_vty_opts->cpu_affinity_li_mutex, NULL);
 
-	install_element(CONFIG_NODE, &cfg_sched_cmd);
+	install_lib_element(CONFIG_NODE, &cfg_sched_cmd);
 	install_node(&sched_node, config_write_sched);
 
-	install_element(L_CPU_SCHED_NODE, &cfg_sched_policy_cmd);
-	install_element(L_CPU_SCHED_NODE, &cfg_sched_cpu_affinity_cmd);
+	install_lib_element(L_CPU_SCHED_NODE, &cfg_sched_policy_cmd);
+	install_lib_element(L_CPU_SCHED_NODE, &cfg_sched_cpu_affinity_cmd);
 
-	install_element_ve(&show_sched_threads_cmd);
+	install_lib_element_ve(&show_sched_threads_cmd);
 
 	/* Initialize amount of cpus now */
 	if (get_num_cpus() < 0)
@@ -631,7 +632,10 @@ int osmo_cpu_sched_vty_apply_localthread(void)
 	int rc = 0;
 
 	/* Assert subsystem was inited and structs are preset */
-	OSMO_ASSERT(sched_vty_opts);
+	if (!sched_vty_opts) {
+		LOGP(DLGLOBAL, LOGL_FATAL, "Setting cpu-affinity mask impossible: no opts!\n");
+		return 0;
+	}
 
 	if (pthread_getname_np(pthread_self(), name, sizeof(name)) == 0)
 		has_name = true;
