@@ -298,7 +298,7 @@ int main(int argc, char **argv)
 
 	osmo_init_logging2(ctx, &info);
 	log_set_use_color(osmo_stderr_target, 0);
-	log_set_print_filename(osmo_stderr_target, 0);
+	log_set_print_filename2(osmo_stderr_target, LOG_FILENAME_NONE);
 
 	msgb_talloc_ctx_init(ctx, 0);
 
