@@ -374,7 +374,7 @@ static void test_create_sapi_reject()
 	struct msgb *msg;
 
 	printf("Testing creating SAPI Reject\n");
-	msg = gsm0808_create_sapi_reject(3);
+	msg = gsm0808_create_sapi_reject_cause(3, GSM0808_CAUSE_BSS_NOT_EQUIPPED);
 	VERIFY(msg, res, ARRAY_SIZE(res));
 	msgb_free(msg);
 }
@@ -499,7 +499,7 @@ static void test_create_ass2()
 		GSM0808_SCT_CSD | 0x90,
 		0xc0,
 		GSM0808_IE_CALL_ID,
-		0xce, 0xfa, 0xad, 0xde, /* CallID */		
+		0xce, 0xfa, 0xad, 0xde, /* CallID */
 		0x83, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, 0x45, /* Kc */
 		GSM0808_IE_GLOBAL_CALL_REF, 0x0d, /* GCR, length */
 		0x03, 0x44, 0x44, 0x44, /* GCR, Net ID */
@@ -1760,6 +1760,48 @@ static void test_gsm0808_enc_dec_cell_id_global()
 	msgb_free(msg);
 }
 
+static void test_gsm0808_enc_dec_cell_id_global_ps()
+{
+	struct gsm0808_cell_id enc_cgi = {
+		.id_discr = CELL_IDENT_WHOLE_GLOBAL,
+		.id.global = {
+			.lai = {
+				.plmn = { .mcc = 123, .mnc = 456 },
+				.lac = 0x2342
+			},
+			.cell_identity = 0x423,
+		}
+	};
+	struct gsm0808_cell_id enc_cgi_ps = {
+		.id_discr = CELL_IDENT_WHOLE_GLOBAL_PS,
+		.id.global_ps = {
+			.rai = {
+				.lac = {
+					.plmn = { .mcc = 123, .mnc = 456 },
+					.lac = 0x2342
+				},
+				.rac = 0xcc,
+			},
+			.cell_identity = 0x423,
+		}
+	};
+	struct msgb *msg_cgi, *msg_cgi_ps;
+	uint8_t rc_enc;
+
+	msg_cgi = msgb_alloc(1024, "output buffer (CGI)");
+	rc_enc = gsm0808_enc_cell_id(msg_cgi, &enc_cgi);
+	OSMO_ASSERT(rc_enc > 0);
+
+	msg_cgi_ps = msgb_alloc(1024, "output buffer (CGI-PS)");
+	rc_enc = gsm0808_enc_cell_id(msg_cgi_ps, &enc_cgi_ps);
+	OSMO_ASSERT(rc_enc > 0);
+
+	OSMO_ASSERT(msgb_eq(msg_cgi, msg_cgi_ps));
+
+	msgb_free(msg_cgi);
+	msgb_free(msg_cgi_ps);
+}
+
 static void test_gsm0808_sc_cfg_from_gsm48_mr_cfg_single(struct gsm48_multi_rate_conf *cfg)
 {
 	uint16_t s15_s0;
@@ -1974,7 +2016,7 @@ static void test_gsm0808_sc_cfg_from_gsm48_mr_cfg(void)
 	cfg.m10_2 = 0;
 	cfg.m12_2 = 0;
 	test_gsm0808_sc_cfg_from_gsm48_mr_cfg_single(&cfg);
-	
+
 }
 
 static void test_gsm48_mr_cfg_from_gsm0808_sc_cfg_single(uint16_t s15_s0)
@@ -2462,6 +2504,7 @@ int main(int argc, char **argv)
 	test_gsm0808_enc_dec_cell_id_ci();
 	test_gsm0808_enc_dec_cell_id_lac_and_ci();
 	test_gsm0808_enc_dec_cell_id_global();
+	test_gsm0808_enc_dec_cell_id_global_ps();
 	test_gsm0808_sc_cfg_from_gsm48_mr_cfg();
 	test_gsm48_mr_cfg_from_gsm0808_sc_cfg();
 
