@@ -37,19 +37,19 @@ static const struct log_info_cat default_categories[] = {
 	[DRLL] = {
 		.name = "DRLL",
 		.description = "A-bis Radio Link Layer (RLL)",
-		.color = "\033[1;31m",
+		.color = OSMO_LOGCOLOR_RED,
 		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
 	[DCC] = {
 		.name = "DCC",
 		.description = "Layer3 Call Control (CC)",
-		.color = "\033[1;32m",
+		.color = OSMO_LOGCOLOR_GREEN,
 		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
 	[DMM] = {
 		.name = NULL,
 		.description = "Layer3 Mobility Management (MM)",
-		.color = "\033[1;33m",
+		.color = OSMO_LOGCOLOR_YELLOW,
 		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
 };
@@ -77,7 +77,8 @@ int main(int argc, char **argv)
 	stderr_target = log_target_create_stderr();
 	log_add_target(stderr_target);
 	log_set_all_filter(stderr_target, 1);
-	log_set_print_filename(stderr_target, 0);
+	log_set_print_filename2(stderr_target, LOG_FILENAME_NONE);
+	log_set_print_category_hex(stderr_target, 0);
 	log_set_print_category(stderr_target, 1);
 	log_set_use_color(stderr_target, 0);
 
